# -*- coding: utf-8 -*-
"""
utils string module.
"""

from decimal import Decimal


def upper(value):
    """
    returns the uppercase copy of input string.

    :param str value: string to make uppercase.

    :rtype: str
    """

    return value.upper()


def lower(value):
    """
    returns the lowercase copy of input string.

    :param str value: string to make lowercase.

    :rtype: str
    """

    return value.lower()


def remove_line_break_escapes(value):
    """
    removes line break escapes from given value.
    it replaces `\\n` with `\n` to enable line breaks.

    :param str value: value to remove line break escapes from it.

    :rtype: str
    """

    return value.replace('\\n', '\n')


def remove_duplicate_space(value):
    """
    removes all duplicate spaces and keeps just a single space in given value.

    :param str value: value to remove duplicate spaces from it.

    :rtype: str
    """

    if value is None:
        return None

    while '  ' in value:
        value = value.replace('  ', ' ')

    return value


def interpolate(message, data):
    """
    interpolates the given message with given data.

    if data is not a dict, no interpolation will be done.

    :param str message: message to be interpolated.
    :param dict data: data to be used for interpolation.

    :rtype: str
    """

    if not isinstance(data, dict):
        return message

    return message.format(**data)


def sort_by_length(items, reverse=False):
    """
    sorts the given list of strings by length of items.

    :param list[str] items: list of strings to be sorted.

    :param bool reverse: sort by descending length.
                         defaults to False if not provided.

    :rtype: list[str]
    """

    return sorted(items, key=len, reverse=reverse)


def quote(value, sign="'"):
    """
    quotes the given string value.

    :param str value: value to be quoted.
    :param str sign: quotation sign to be used.
                     defaults to single quotation if not provided.

    :rtype: str
    """

    return "{sign}{value}{sign}".format(sign=sign, value=str(value))


def union(first, second, **options):
    """
    gets the list of strings which are in both lists.

    note that it returns the value from the first list.
    by default comparison is case-insensitive.
    so if `A` is in the first list and `a` is in the second list, the
    result list will contain `A` from the first list.

    :param list[str] | set[str] | tuple[str] first: first list of strings.
    :param list[str] | set[str] | tuple[str] second: second list of strings.

    :keyword function converter: a callable to be used for case conversion of
                                 strings. it must take a single input string and
                                 return the converted value. if you want case-sensitive
                                 comparison, you could pass `str` function as converter.
                                 defaults to `lower` which results in case-insensitive
                                 comparison.

    :keyword type collection: collection type to be used for result.
                              defaults to `list` if not provided.

    :rtype: list[str] | set[str] | tuple[str]
    """

    collection = options.get('collection', list)
    converter = options.get('converter', lower)
    result = []
    second = [converter(item) for item in second]
    for item in first:
        if converter(item) in second:
            result.append(item)

    return collection(result)


def is_match(item, values, **options):
    """
    gets a value indicating that given item exists in the given list of values.

    by default comparison is case-insensitive.
    so if item is `A` and `a` is in the values list, it returns True.

    :param str item: value to be checked for matching.
    :param list[str] | set[str] | tuple[str] values: list of string values.

    :keyword function converter: a callable to be used for case conversion of
                                 strings. it must take a single input string and
                                 return the converted value. if you want case-sensitive
                                 comparison, you could pass `str` function as converter.
                                 defaults to `lower` which results in case-insensitive
                                 comparison.

    :rtype: bool
    """

    result = union([item], values, **options)
    return len(result) > 0


def coerce_to_string(value, *accepted_types):
    """
    gets the string representation of given value.

    it only converts value to string if it is of provided types.
    if no type is given, it only converts integers, floats and decimals.
    otherwise returns the same value.

    :param object value: value to be converted.

    :param type accepted_types: accepted types for conversion to string.
                                if not provided, defaults to (int, float, Decimal) types.

    :returns: string or the same value.
    """

    if bool not in accepted_types and isinstance(value, bool):
        return value

    if accepted_types is None or len(accepted_types) <= 0:
        accepted_types = (int, float, Decimal)

    if isinstance(value, accepted_types):
        return str(value)

    return value



# -*- coding: utf-8 -*-
"""
cors package.
"""

from pyrin.packaging.base import Package


class CORSPackage(Package):
    """
    cors package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration']
    CONFIG_STORE_NAMES = ['cors']
    COMPONENT_NAME = 'processor.cors.component'



# -*- coding: utf-8 -*-
"""
caching component module.
"""

from pyrin.application.decorators import component
from pyrin.caching import CachingPackage
from pyrin.caching.manager import CachingManager
from pyrin.application.structs import Component


@component(CachingPackage.COMPONENT_NAME)
class CachingComponent(Component, CachingManager):
    """
    caching component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
aes128 encryption handler module.
"""

from cryptography.fernet import Fernet

import pyrin.configuration.services as config_services

from pyrin.security.encryption.decorators import encrypter
from pyrin.security.encryption.handlers.base import SymmetricEncrypterBase


@encrypter()
class AES128Encrypter(SymmetricEncrypterBase):
    """
    aes128 encrypter class.
    """

    def __init__(self, **options):
        """
        initializes an instance of AES128Encrypter.
        """

        super().__init__(**options)

        self._encrypter = Fernet(self._get_encryption_key(**options))

    def _get_encryption_key(self, **options):
        """
        gets the signing key for encryption.

        :rtype: str
        """

        return config_services.get('security', 'encryption', 'aes128_key')

    def _get_algorithm(self, **options):
        """
        gets the algorithm used for encryption.

        :rtype: str
        """

        return 'AES128'

    def _encrypt(self, text, **options):
        """
        encrypts the given value and returns the encrypted result.

        :param str text: text to be encrypted.

        :rtype: bytes
        """

        return self._encrypter.encrypt(text.encode(self._encoding))

    def _decrypt(self, value, **options):
        """
        decrypts the given value and returns the decrypted result.

        :param bytes value: value to be decrypted.

        :rtype: str
        """

        return self._encrypter.decrypt(value).decode(self._encoding)

    def generate_key(self, **options):
        """
        generates a valid key for this handler and returns it.

        :rtype: str
        """

        return Fernet.generate_key().decode(self._encoding)



# -*- coding: utf-8 -*-
"""
hashing exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class HashingManagerException(CoreException):
    """
    hashing manager exception.
    """
    pass


class HashingManagerBusinessException(CoreBusinessException,
                                      HashingManagerException):
    """
    hashing manager business exception.
    """
    pass


class InvalidHashingHandlerTypeError(HashingManagerException):
    """
    invalid hashing handler type error.
    """
    pass


class DuplicatedHashingHandlerError(HashingManagerException):
    """
    duplicated hashing handler error.
    """
    pass


class HashingHandlerNotFoundError(HashingManagerException):
    """
    hashing handler not found error.
    """
    pass


class InvalidHashingHandlerNameError(HashingManagerException):
    """
    invalid hashing handler name error.
    """
    pass


class InvalidHashError(HashingManagerBusinessException):
    """
    invalid hash error.
    """
    pass



# -*- coding: utf-8 -*-
"""
configuration services module.
"""

from pyrin.application.services import get_component
from pyrin.configuration import ConfigurationPackage


def load_configuration(name, **options):
    """
    loads the given configuration if relevant file is available in settings path.

    it creates the config file in application settings path if
    config file is only available in pyrin default settings.

    :param str name: configuration name.

    :keyword dict defaults: a dict containing values
                            needed for interpolation.
                            defaults to None if not provided.

    :keyword bool silent: specifies that if a related configuration file
                          for the given name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword bool ignore_on_existed: specifies that it should not raise an
                                     error if a config store with given name
                                     has been already loaded.
                                     defaults to False if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationStoreExistedError: configuration store existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).load_configuration(name,
                                                                                 **options)


def load_configurations(*names, **options):
    """
    loads the given configurations if relevant files is available in settings path.

    it creates the config files in application settings path if
    config files are only available in pyrin default settings.

    :param str names: configuration names as arguments.

    :keyword dict defaults: a dict containing values
                            needed for interpolation.
                            defaults to None if not provided.

    :keyword bool silent: specifies that if a related configuration file
                          for any of the given names not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword bool ignore_on_existed: specifies that it should not raise an
                                     error if a config store with given name
                                     has been already loaded.
                                     defaults to False if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationStoreExistedError: configuration store existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).load_configurations(*names,
                                                                                  **options)


def reload(store_name, **options):
    """
    reloads the configuration store from it's relevant file.

    :param str store_name: config store name to be reloaded.

    :keyword dict defaults: a dict containing values
                            needed for interpolation.
                            defaults to None if not provided.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).reload(store_name, **options)


def get_file_path(store_name, **options):
    """
    gets the configuration file path for given config store.

    :param str store_name: config store name to get its file path.

    :keyword bool silent: specifies that if a related configuration file
                          for the given store name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_file_path(store_name,
                                                                            **options)


def get_default_file_path(store_name, **options):
    """
    gets the default configuration file path for given config store.

    it gets the file path in `pyrin.settings.default` directory.
    if the file is not available, it could return None or raise an error.

    :param str store_name: config store name to get its file path.

    :keyword bool silent: specifies that if a related default configuration
                          file for the given store name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_default_file_path(
        store_name, **options)


def get_file_name(store_name, **options):
    """
    gets the configuration file name for given config store.

    for example: `database.ini`

    :param str store_name: config store name to get its file name.

    :keyword bool silent: specifies that if a related configuration file
                          for the given store name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_file_name(store_name,
                                                                            **options)


def create_config_files(*stores, **options):
    """
    creates the config files for given store names in application settings path.

    it creates the files based on files available in pyrin default settings.

    :param str stores: store names to create config file for them.

    :keyword bool replace_existing: specifies that it should replace file if a config
                                    file is already available for given store name.
                                    this argument takes precedence over
                                    `ignore_on_existed` argument.
                                    defaults to False if not provided.

    :keyword bool ignore_on_existed: specifies that if a config file for the
                                     store name is already existed, ignore
                                     it, otherwise raise an error.
                                     defaults to False if not provided.

    :keyword bool silent: specifies that if a related default configuration
                          file for the given store name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword dict[str, str] data: keyword with values to be replaced in config file.
                                  any keyword that does not exist in config file, will
                                  be ignored. defaults to None if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationFileExistedError: configuration file existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).create_config_files(*stores,
                                                                                  **options)


def create_config_file(store, **options):
    """
    creates the config file for given store name in application settings path.

    it creates the file based on file available in pyrin default settings.

    :param str store: store name to create config file for it.

    :keyword bool replace_existing: specifies that it should replace file if a config
                                    file is already available for given store name.
                                    this argument takes precedence over
                                    `ignore_on_existed` argument.
                                    defaults to False if not provided.

    :keyword bool ignore_on_existed: specifies that if a config file for the
                                     store name is already existed, ignore
                                     it, otherwise raise an error.
                                     defaults to False if not provided.

    :keyword bool silent: specifies that if a related default configuration
                          file for the given store name not found, ignore it.
                          otherwise raise an error. defaults to False.

    :keyword dict[str, str] data: keyword with values to be replaced in config file.
                                  any keyword that does not exist in config file, will
                                  be ignored. defaults to None if not provided.

    :raises ConfigurationFileNotFoundError: configuration file not found error.
    :raises ConfigurationFileExistedError: configuration file existed error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).create_config_file(store,
                                                                                 **options)


def get(store_name, section, key, **options):
    """
    gets the value of specified key from provided section of given config store.

    :param str store_name: config store name.
    :param str section: config section name.
    :param str key: config key to get it's value.

    :keyword object default: default value if key not present in config section.
                             if not provided, error will be raised.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get(store_name, section,
                                                                  key, **options)


def get_active(store_name, key, **options):
    """
    gets the value of given key from active section of given config store.

    if this store does not have an active section, it raises an error.

    :param str store_name: config store name.
    :param str key: config key to get it's value.

    :keyword object default: default value if key not present in config section.
                             if not provided, error will be raised.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active(store_name,
                                                                         key, **options)


def get_section_names(store_name, **options):
    """
    gets all available section names of given config store.

    :param str store_name: config store name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :rtype: list[str]
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section_names(store_name,
                                                                                **options)


def get_section(store_name, section, **options):
    """
    gets all key/values stored in given section of specified config store.

    :param str store_name: config store name.
    :param str section: section name.

    :keyword callable converter: a callable to use as case converter for keys.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section(store_name, section,
                                                                          **options)


def get_section_keys(store_name, section, **options):
    """
    gets all available keys in given section of specified config store.

    :param str store_name: config store name.
    :param str section: section name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :rtype: list[str]
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_section_keys(store_name,
                                                                               section,
                                                                               **options)


def get_all(store_name, **options):
    """
    gets all available key/values from different sections of given config store.

    in a flat dict, eliminating the sections.
    note that if there are same key names with different values
    in different sections, it raises an error to prevent overwriting
    values. also note that if given config store contains `active` section,
    then the result of `get_active_section` method would be returned.

    :param str store_name: config store name.

    :keyword callable converter: a callable to use as case converter for keys.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.
    :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_all(store_name, **options)


def get_active_section(store_name, **options):
    """
    gets the active section available in given config store.

    this method gets the section that it's name is under [active]
    section, for example:

    [active]
    selected: production

    [production]
    id: 123
    name: prod

    [development]
    id: 233
    name: dev

    this will return all key/values available under [production].
    if the config store has not an [active] section, this method
    raises an error.

    :param str store_name: config store name.

    :keyword callable converter: a callable to use as case converter.
                                 it should be a callable with a signature
                                 similar to below example:
                                 case_converter(input_dict).

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                    not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active_section(store_name,
                                                                                 **options)


def get_active_section_name(store_name):
    """
    gets the active section name of given config store if available.

    if the store does not have an active section, it raises an error.

    :param str store_name: config store name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                    section not found error.

    :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                key not found error.

    :rtype: str
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_active_section_name(store_name)


def get_all_sections(store_name, **options):
    """
    gets all sections and their keys of given config store.

    :param str store_name: config store name.

    :raises ConfigurationStoreNotFoundError: configuration store not found error.

    :rtype: dict
    """

    return get_component(ConfigurationPackage.COMPONENT_NAME).get_all_sections(store_name,
                                                                               **options)



# -*- coding: utf-8 -*-
"""
logging component module.
"""

from pyrin.application.decorators import component
from pyrin.logging import LoggingPackage
from pyrin.logging.manager import LoggingManager
from pyrin.application.structs import Component


@component(LoggingPackage.COMPONENT_NAME)
class LoggingComponent(Component, LoggingManager):
    """
    logging component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
template handlers base module.
"""

import os

from abc import abstractmethod

import pyrin.configuration.services as config_services
import pyrin.utils.path as path_utils
import pyrin.utils.file as file_utils

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.utils.custom_print import print_default
from pyrin.template.interface import AbstractTemplateHandler
from pyrin.template.exceptions import TemplateHandlerNameRequiredError, \
    InvalidSourceDirectoryError, InvalidTargetDirectoryError, \
    TemplateTargetDirectoryAlreadyExistedError


class TemplateHandlerBase(AbstractTemplateHandler):
    """
    template handler base class.

    all template handlers must be subclassed from this.
    """

    def __init__(self, name, source, target, **options):
        """
        initializes an instance of TemplateHandlerBase.

        :param str name: name of the handler.
                         each handler must have a unique name.

        :param str source: source directory of template files.

        :param str target: target directory in which template
                           files must be created in.
                           it could be set to None to be calculated
                           in `_create()` method based on given inputs.

        :raises TemplateHandlerNameRequiredError: template handler name required error.
        :raises InvalidSourceDirectoryError: invalid source directory error.
        """

        super().__init__()

        if name in (None, '') or name.isspace():
            raise TemplateHandlerNameRequiredError('Template handler name is required.')

        if source in (None, '') or source.isspace() or \
                not os.path.isabs(source) or not os.path.exists(source) or \
                not os.path.isdir(source):
            raise InvalidSourceDirectoryError('The specified source directory is invalid.')

        self._name = name
        self._source = source
        self._target = target
        self._config_stores = self._get_config_stores()
        self._data = None
        self._config_data = None

    def create(self, *args, **kwargs):
        """
        creates the template files in target path of this handler.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.

        :raises InvalidTargetDirectoryError: invalid target directory error.

        :raises TemplateTargetDirectoryAlreadyExistedError: template target directory
                                                            already existed error.
        """

        self._create(*args, **kwargs)
        self._data = self._get_data()
        self._config_data = self._get_config_data()
        self._validate_target()
        self._copy_config_files()
        self._copy_template_files()
        self._replace_template_values()
        self._create_required_directories()
        self._finalize()

        self._print_info('Template has been created in [{target}].'
                         .format(target=self._target))

    def _create(self, *args, **kwargs):
        """
        creates the template files in target path of this handler.

        this method is intended to be overridden by subclasses.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.
        """
        pass

    def _validate_target(self):
        """
        validates target path for template creation.

        it assures that target path does not exist, but subclasses could
        override this method if different type of validation is required.

        :raises InvalidTargetDirectoryError: invalid target directory error.

        :raises TemplateTargetDirectoryAlreadyExistedError: template target directory
                                                            already existed error.
        """

        if self._target in (None, '') or self._target.isspace() or \
                not os.path.isabs(self._target):
            raise InvalidTargetDirectoryError('The specified target directory is invalid.')

        if os.path.exists(self._target):
            raise TemplateTargetDirectoryAlreadyExistedError('Target directory [{dir}] is '
                                                             'already existed. if you want '
                                                             'to regenerate this template, '
                                                             'you should delete the existing '
                                                             'directory first.'
                                                             .format(dir=self._target))

    def _copy_config_files(self):
        """
        copies any required config files for current template handler.
        """

        if len(self._config_stores) > 0:
            config_services.create_config_files(*self._config_stores,
                                                replace_existing=True,
                                                data=self._config_data)

    def _copy_template_files(self):
        """
        copies required template files from source path to target path of this handler.
        """

        path_utils.copy_directory(self._source, self._target,
                                  ignore_existed=True, ignore=path_utils.get_pycache)

    def _replace_template_values(self):
        """
        replaces the values of files in target directory with values in given dict.
        """

        if len(self._data):
            file_utils.replace_files_values(self._target, self._data,
                                            *self._get_file_patterns())

    def _create_required_directories(self):
        """
        creates the required directories if they are not available in target path.
        """

        if len(self._get_required_directories()) > 0:
            for directory in self._get_required_directories():
                full_path = os.path.abspath(os.path.join(self._target, directory))
                if not os.path.exists(full_path):
                    path_utils.create_directory(full_path)

    def _get_file_patterns(self):
        """
        gets the file patterns that should be included in replacement operation.

        this method is intended to be overridden by subclasses.

        :rtype: list[str]
        """

        return []

    def _get_data(self):
        """
        gets the data required in template generation to replace in files.

        this method is intended to be overridden by subclasses.

        :rtype: dict
        """

        return {}

    def _get_config_data(self):
        """
        gets the data required in template generation to replace in config files.

        this method is intended to be overridden by subclasses.

        :rtype: dict
        """

        return {}

    def _get_config_stores(self):
        """
        gets the config store names which belong to this template.

        this method is intended to be overridden by subclasses.

        :rtype: list[str]
        """

        return []

    def _get_required_directories(self):
        """
        gets the required directory names that must be created in target path.

        this is useful for directories which are in template files but are empty
        and could not be included in `setup.py` package data.

        :rtype: list[str]
        """

        return []

    def _finalize(self):
        """
        finalizes the template creation.

        this method is intended to be overridden by subclasses.
        """
        pass

    def _print_info(self, message):
        """
        prints the message as info.

        this method could be overridden in subclasses if
        printing is not desired for a template handler.

        :param str message: message to be printed.
        """

        print_default(message, force=True)

    @property
    def name(self):
        """
        gets the name of this template handler.

        :rtype: str
        """

        return self._name


class TemplateHandlerWithInputBase(TemplateHandlerBase):
    """
    template handler with input base class.

    this type of template handler will get the inputs
    from arguments passed in cli command.
    """

    def __init__(self, name, source, **options):
        """
        initializes an instance of TemplateHandlerWithInputBase.

        :param str name: name of the handler.
                         each handler must have a unique name.

        :param str source: source directory of template files.

        :raises TemplateHandlerNameRequiredError: template handler name required error.
        :raises InvalidSourceDirectoryError: invalid source directory error.
        """

        super().__init__(name, source, None, **options)

    def _create(self, *args, **kwargs):
        """
        creates the template files in target path of this handler.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.
        """

        self._process_inputs(*args, **kwargs)

    @abstractmethod
    def _process_inputs(self, *args, **kwargs):
        """
        processes the inputs to be used by this handler.

        this method is intended to be overridden by subclasses.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()


class TemplateHandlerWithInterfaceInputBase(TemplateHandlerWithInputBase):
    """
    template handler with interface input base class.

    this type of template handler will show cli prompt to user to get inputs.
    it also could get the inputs from arguments passed in cli command.
    """

    def _process_inputs(self, *args, **kwargs):
        """
        processes the inputs to be used by this handler.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.
        """

        self._show_interface(*args, **kwargs)

    @abstractmethod
    def _show_interface(self, *args, **kwargs):
        """
        shows cli prompt to get inputs from user.

        this method is intended to be overridden by subclasses.

        :param object args: arguments passed through command line.

        :keyword object kwargs: keyword arguments passed through command line.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
request wrappers structs module.
"""

from pyrin.core.structs import Context


class RequestContext(Context):
    """
    context class to hold request contextual data.
    """

    attribute_error_message = 'Property [{name}] not found in request context.'



# -*- coding: utf-8 -*-
"""
datetime services module.
"""

from pyrin.application.services import get_component
from pyrin.globalization.datetime import DateTimePackage


def now(server=True, timezone=None):
    """
    gets the current datetime based on requested timezone.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get datetime based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).now(server, timezone)


def client_now():
    """
    gets the current datetime based on client timezone.

    this is a helper method to let get the client datetime
    without providing value to `now` method.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).client_now()


def current_year(server=True, timezone=None):
    """
    gets the current year based on requested timezone.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get year based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: int
    """

    return get_component(DateTimePackage.COMPONENT_NAME).current_year(server, timezone)


def current_client_year():
    """
    gets the current year based on client timezone.

    this is a helper method to let get the client year
    without providing value to `current_year` method.

    :rtype: int
    """

    return get_component(DateTimePackage.COMPONENT_NAME).current_client_year()


def get_default_client_timezone():
    """
    gets the default client timezone.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_default_client_timezone()


def get_current_timezone(server):
    """
    gets the current timezone for server or client.

    :param bool server: if set to True, server timezone will be returned.
                        if set to False, client timezone will be returned.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timezone(server)


def get_timezone(timezone):
    """
    gets the timezone based on given timezone name.

    :param str timezone: timezone name.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_timezone(timezone)


def get_timezone_name(server):
    """
    gets the server or client timezone name.

    :param bool server: if set to True, server timezone name will be returned.
                        if set to False, client timezone name will be returned.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_timezone_name(server)


def as_timezone(value, server):
    """
    gets the result of `astimezone` on the given value.

    :param datetime value: value to get normalized.

    :param bool server: if set to True, server timezone name will be used.
                        if set to False, client timezone name will be used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).as_timezone(value, server)


def normalize(value, server):
    """
    normalizes input value using server or client current timezone and returns it.

    :param datetime value: value to get normalized.
    :param bool server: specifies that server or client timezone must used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).normalize(value, server)


def localize(value, server):
    """
    localizes input datetime with current timezone.

    input value should not have a timezone info.

    :param datetime value: value to be localized.
    :param bool server: specifies that server or client timezone must used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).localize(value, server)


def replace_timezone(value, server):
    """
    replaces given value's timezone with server or client timezone.

    it returns a new object.
    note that this method does not normalize the value, it just replaces the
    timezone and localizes the value.

    :param datetime value: value to replace its timezone.
    :param bool server: specifies that server or client timezone must used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).replace_timezone(value, server)


def to_datetime_string(value, server):
    """
    gets the datetime string representation of input value.

    if the value has no timezone info, it adds the client or server
    timezone info based on `server` value.

    example: `2015-12-24T22:40:15+01:00`

    :param datetime value: input object to be converted.

    :param bool server: specifies that value must be normalized
                        to server or client timezone.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime_string(value, server)


def to_date_string(value):
    """
    gets the date string representation of input value.

    example: `2015-12-24`

    :param datetime | date value: input object to be converted.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date_string(value)


def to_time_string(value, server):
    """
    gets the time string representation of input value.

    if the value is a datetime and has no timezone info, it adds
    the client or server timezone info based on `server` value.

    example: `23:40:15`

    :param datetime | time value: input object to be converted.

    :param bool server: specifies that value must be normalized
                        to server or client timezone.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time_string(value, server)


def to_datetime(value, server, replace_server=None):
    """
    converts the input value to it's equivalent python datetime.

    :param str value: string representation of datetime to be converted.

    :param bool server: specifies that value must be normalized
                        to server or client timezone.

    :param bool replace_server: specifies that it must replace the timezone
                                of value with timezone of server before
                                normalization. if set to False, it replaces
                                it with client timezone.
                                defaults to None and no replacement will be done.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime(value, server,
                                                                     replace_server)


def to_date(value):
    """
    converts the input value to it's equivalent python date.

    :param str value: string representation of date to be converted.

    :rtype: date
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date(value)


def to_time(value):
    """
    converts the input value to it's equivalent python time.

    :param str value: string representation of time to be converted.

    :rtype: time
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time(value)


def timezone_exists(timezone_name):
    """
    gets a value indicating that a timezone with the given name exists.

    :param str timezone_name: timezone name to check for existence.

    :rtype: bool
    """

    return get_component(DateTimePackage.COMPONENT_NAME).timezone_exists(timezone_name)


def get_current_timestamp(date_sep='-', main_sep=' ',
                          time_sep=':', server=True, timezone=None):
    """
    gets the current timestamp with specified separators based on requested timezone.

    :param str date_sep: a separator to put between date elements.
    :param str main_sep: a separator to put between date and time part.
    :param str time_sep: a separator to put between time elements.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get datetime based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timestamp(date_sep,
                                                                               main_sep,
                                                                               time_sep,
                                                                               server,
                                                                               timezone)



# -*- coding: utf-8 -*-
"""
router handlers protected module.
"""

import pyrin.security.authorization.services as authorization_services
import pyrin.security.session.services as session_services
import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _
from pyrin.api.router.handlers.base import RouteBase, TemporaryRouteBase
from pyrin.processor.request.enumerations import RequestHeaderEnum
from pyrin.security.permission.base import PermissionBase
from pyrin.api.router.handlers.exceptions import FreshAuthenticationRequiredError, \
    PermissionTypeError


class ProtectedRoute(RouteBase):
    """
    protected route class.

    this class should be used to manage application protected api
    routes that require authenticated access.
    """

    def __init__(self, rule, **options):
        """
        initializes an instance of ProtectedRoute.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls and http methods will be
                         overwritten if `replace=True` option is provided.
                         otherwise an error will be raised.

        :keyword str | tuple[str] methods: http methods that this route could handle.
                                           if not provided, defaults to `GET`, `HEAD`
                                           and `OPTIONS`.

        :keyword function view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword ResultSchema result_schema: result schema to be used to filter results.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided. it will
                                                      be used only for entity conversion. this
                                                      value will override the corresponding
                                                      value of `result_schema` if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    defaults to False if not provided.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                     to access this route.

        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.
        :raises PermissionTypeError: permission type error.
        """

        super().__init__(rule, **options)

        if self._cors is not None:
            self._cors.add_allowed_headers(RequestHeaderEnum.AUTHORIZATION)

        self._permissions = options.get('permissions', None)
        self._permissions = misc_utils.make_iterable(self._permissions, tuple)

        if not all(isinstance(item, PermissionBase) for item in self._permissions):
            raise PermissionTypeError('All route permissions must be an '
                                      'instance of [{instance}].'
                                      .format(instance=PermissionBase))

    def _handle(self, inputs, **options):
        """
        handles the current route.

        routes which need to perform extra operations before
        view function execution, must override this method.

        :param dict inputs: view function inputs.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        self._authorize()
        super()._handle(inputs, **options)

    def _authorize(self):
        """
        authorizes the route permissions for current user.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        user = session_services.get_current_user()
        authorization_services.authorize(user, self.permissions)

    @property
    def permissions(self):
        """
        gets all required permissions to access this route.

        :returns: tuple[PermissionBase]

        :rtype: tuple
        """

        return self._permissions


class FreshProtectedRoute(ProtectedRoute):
    """
    fresh protected route class.

    this class should be used to manage application protected api routes that
    require fresh authenticated access. fresh authentication means an authentication
    that has been done by providing user credentials to server.
    """

    def _authorize(self):
        """
        authorizes the route permissions for current user.

        also checks that the user has a fresh authentication.

        :raises FreshAuthenticationRequiredError: fresh authentication required error.
        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        if not session_services.is_fresh():
            raise FreshAuthenticationRequiredError(_('Fresh authentication is required '
                                                     'to access the requested resource.'))

        super()._authorize()


class ProtectedTemporaryRoute(ProtectedRoute, TemporaryRouteBase):
    """
    protected temporary route class.

    this class should be used to manage application protected api
    routes that require authenticated access and must be unregistered
    after a specified number of requests or a duration of time.
    this is useful for a couple of different things but mostly for system
    health checking after each new deployment on production.
    """
    pass


class FreshProtectedTemporaryRoute(FreshProtectedRoute, TemporaryRouteBase):
    """
    fresh protected temporary route class.

    this class should be used to manage application protected api routes
    that require fresh authenticated access and must be unregistered
    after a specified number of requests or a duration of time.
    this is useful for a couple of different things but mostly for system
    health checking after each new deployment on production.
    """
    pass



# -*- coding: utf-8 -*-
"""
core structs module.
"""

from collections import deque
from threading import Lock
from abc import abstractmethod

from werkzeug.datastructures import MultiDict, ImmutableMultiDict, ImmutableDict, Headers

import pyrin.utils.misc as misc_utils

from pyrin.core.exceptions import CoreAttributeError, ContextAttributeError, \
    CoreNotImplementedError, PackageClassIsNotSetError, CoreKeyError


class SingletonMetaBase(type):
    """
    singleton meta base class.

    this is a thread-safe implementation of singleton.
    """

    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        if cls._has_instance() is False:
            with cls._lock:
                if cls._has_instance() is False:
                    instance = super().__call__(*args, **kwargs)
                    cls._register_instance(instance)

        return cls._get_instance()

    @abstractmethod
    def _has_instance(cls):
        """
        gets a value indicating there is a registered instance.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _register_instance(cls, instance):
        """
        registers the given instance.

        :param object instance: instance to be registered.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_instance(cls):
        """
        gets the registered instance.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()


class UniqueSingletonMeta(SingletonMetaBase):
    """
    unique singleton meta class.

    this is a thread-safe implementation of singleton.
    this class only allows a single unique object for all descendant types.

    for example: {Base -> UniqueSingletonMeta, A -> Base, B -> A}
    if some_object = Base() then always Base() = A() = B() = some_object.
    or if some_object = A() then always A() = B() = some_object != Base().
    """

    _instance = None
    _lock = Lock()

    def _has_instance(cls):
        """
        gets a value indicating that there is a registered instance.

        :rtype: bool
        """

        return cls._instance is not None

    def _register_instance(cls, instance):
        """
        registers the given instance.
        """

        cls._instance = instance

    def _get_instance(cls):
        """
        gets the registered instance.

        :rtype: object
        """

        return cls._instance


class MultiSingletonMeta(SingletonMetaBase):
    """
    multi singleton meta class.

    this is a thread-safe implementation of singleton.
    this class allows a unique object per each type of descendants.

    for example: {Base -> UniqueSingletonMeta, A -> Base, B -> A}
    if some_object = Base() then always Base() != A() != B() but always Base() = some_object.
    or if some_object = A() then always Base() != A() != B() but always A() = some_object.
    """

    # a dictionary containing an instance of each type.
    # in the form of: {type: instance}
    _instances = dict()
    _lock = Lock()

    def _has_instance(cls):
        """
        gets a value indicating that there is a registered instance.

        :rtype: bool
        """

        return cls in cls._instances

    def _register_instance(cls, instance):
        """
        registers the given instance.
        """

        cls._instances[cls] = instance

    def _get_instance(cls):
        """
        gets the registered instance.

        :rtype: object
        """

        return cls._instances.get(cls)


class DTO(dict):
    """
    context class for storing objects in every layer.

    it's actually a dictionary with the capability to treat keys as instance attributes.
    """

    def __getattr__(self, name):
        if name in self:
            return self.get(name)

        raise CoreAttributeError('Property [{name}] not found.'.format(name=name))

    def __getitem__(self, item):
        if item in self:
            return self.get(item)

        raise CoreKeyError('Key [{name}] not found.'.format(name=item))

    def __setattr__(self, name, value):
        self[name] = value


class CoreObject(object):
    """
    core object class.

    this should be used as the base object for all application objects.
    """

    def __init__(self):
        """
        initializes an instance of CoreObject.
        """

        super().__init__()
        self.__name = None

    def __setattr__(self, name, value):
        return self._setattr(name, value)

    def __repr__(self):
        """
        gets the string representation of current object.

        :rtype: str
        """

        return str(self)

    def __str__(self):
        """
        gets the string representation of current object.

        :rtype: str
        """

        return self.get_fully_qualified_name()

    def get_name(self):
        """
        gets the name of the object.

        if name is not available, returns its class name.

        :rtype: str
        """

        if self.__name is not None:
            return self.__name
        return self.get_class_name()

    def _set_name(self, name):
        """
        sets new name to current object.

        :param str name: object new name.
        """

        self.__name = name

    def get_class_name(self):
        """
        gets the object's class name.

        :rtype: str
        """

        return self.__class__.__name__

    def get_module_name(self):
        """
        gets the object's module name.

        :rtype: str
        """

        return self.__class__.__module__

    def get_doc(self):
        """
        gets the docstring of the object.

        :rtype: str
        """

        return self.__doc__

    def get_fully_qualified_name(self):
        """
        gets fully qualified name of this object.

        it gets `module_name.class_name` as fully qualified name.

        :rtype: str
        """

        return '{module}.{name}'.format(module=self.__module__,
                                        name=self.__class__.__name__)

    def _setattr(self, name, value):
        """
        sets the given value to specified attribute.

        :param str name: attribute name.
        :param object value: attribute value.
        """

        return super().__setattr__(name, value)


class Context(DTO):
    """
    context class for storing objects in every layer.

    it's actually a dictionary with the capability to add keys directly.
    """

    attribute_error = ContextAttributeError
    attribute_error_message = 'Property [{name}] not found.'

    def __getattr__(self, name):
        if name in self:
            return self.get(name)

        self._raise_key_error(name)

    def __getitem__(self, item):
        if item in self:
            return self.get(item)

        self._raise_key_error(item)

    def _raise_key_error(self, key):
        """
        raises an error for given key.

        :param object key: key object that caused the error.

        :raises ContextAttributeError: context attribute error.
        """

        raise self.attribute_error(self.attribute_error_message.format(name=key))


class HookSingletonMeta(MultiSingletonMeta):
    """
    hook singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class Hook(CoreObject, metaclass=HookSingletonMeta):
    """
    base hook class.

    all application hook classes must be subclassed from this one.
    """
    pass


class ManagerSingletonMeta(MultiSingletonMeta):
    """
    manager singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class Manager(CoreObject, metaclass=ManagerSingletonMeta):
    """
    base manager class.

    all application manager classes must be subclassed from this one.
    """

    # this attribute should be set with the package class of current manager.
    # this is useful if you want to extend pyrin packages in your application
    # and let pyrin use your custom package's package class in its code.
    package_class = None

    def get_package_class(self):
        """
        gets the package class of current manager.

        this method is useful if you want to access the correct package class of a
        manager using services module of that package.

        each package that needs to expose this method, could implement a service method
        and return the result of this method.

        :raises PackageClassIsNotSetError: package class is not set error.

        :returns: type[Package]
        """

        if self.package_class is None:
            raise PackageClassIsNotSetError('Package class for current manager '
                                            '[{manager}] is not set. you must set '
                                            '"package_class" attribute for this manager '
                                            'to be able to use this method.'
                                            .format(manager=self))

        return self.package_class


class CLISingletonMeta(MultiSingletonMeta):
    """
    cli singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class CLI(CoreObject, metaclass=CLISingletonMeta):
    """
    base cli class.

    all application cli classes must be subclassed from this one.
    """

    # this value must be set in each subclass with the relevant callable
    # execute service with the param signature of: `(str handler_name, **inputs)`
    _execute_service = None

    @classmethod
    def execute(cls, handler_name, **options):
        """
        executes the handler with the given name with given inputs.

        :param str handler_name: handler name to be executed.

        :raises CLIHandlerNotFoundError: cli handler not found error.

        :rtype: int
        """

        return cls._execute_service(handler_name, **options)


class Stack(deque):
    """
    stack class.

    this class extends `deque` and provides a `peek()` method to
    just get the top most item without removing it.
    it also provides some other useful methods for convenient of usage.

    note that `Stack` is not guaranteed to be thread-safe on all python
    implementations, because it extends `deque`. so do not use it when
    there is a multi-thread access to the same stack.
    """

    def peek(self):
        """
        gets the top most item of stack, without removing it.

        the return value of `peek()` is the same as `pop()`
        but without removing it from the stack.
        if the stack is empty, it raises an error.

        :raises IndexError: index error.

        :rtype: object
        """

        return self[-1]

    def peek_all(self):
        """
        gets all items of stack, without removing them.

        the result is a list of all items of the stack in the order of
        `pop()`. meaning that, last inserted items will be in lower indices.

        :rtype: list[object]
        """

        return list(reversed(self))

    def push(self, value):
        """
        adds the given value into top of stack.

        this method is just implemented for convenient of usage,
        it will call `append()` under the hood.

        :param object value: value to be added into stack.
        """

        self.append(value)

    def dispose(self):
        """
        deletes the top most item of stack, without returning it.

        if the stack is empty, it raises an error.

        :raises IndexError: index error.
        """

        del self[-1]


class CoreMultiDict(MultiDict):
    """
    core multi dict class.

    this class extends MultiDict to let `to_dict` method gets values that are
    not multiple as a single object instead of a list with a single item.
    """

    def to_dict(self, flat=True, **options):
        """
        returns the contents as regular dict.

        if `flat=True` the returned dict will only have the first item present.

        if `flat=False` and `all_list=True` is provided, all values will be get as list.

        if `flat=False` and `all_list=False` is provided, all values that are multiple
        will be get as list. but single values will be get as a single object.

        :param bool flat: if set to False the dict returned will have lists
                          with all the values in it. otherwise it will only
                          contain the first value for each key.
                          defaults to True if not provided.

        :keyword bool all_list: if set to True and `flat` is set to False, all
                                values will be returned as list. but if set to
                                False and `flat` is also set to False, values
                                that are multiple will be returned as list and
                                single values will be returned as a single object.
                                if `flat` is set to True, this parameter has no
                                effect. defaults to True if not provided.

        :rtype: dict
        """

        all_list = options.get('all_list')
        if all_list is None:
            all_list = True

        if flat is True or all_list is True:
            return super().to_dict(flat=flat)

        return dict(self.singles_or_lists())

    def singles_or_lists(self):
        """
        returns an iterator of `(key, values)` pairs.

        where values is the list of all values associated with the key or
        a single object if values have only a single item in it.
        """

        for key, values in misc_utils.iterate_items(dict, self):
            if len(values) == 1:
                yield key, values[0]
            else:
                yield key, list(values)


class CoreImmutableMultiDict(ImmutableMultiDict, CoreMultiDict):
    """
    core immutable multi dict class.
    """

    def copy(self):
        """
        returns a shallow mutable copy of this object.

        keep in mind that the standard library's `copy` function is a
        no-op for this class like for any other python immutable type.
        """

        return CoreMultiDict(self)


class CoreImmutableDict(ImmutableDict):
    """
    core immutable dict class.
    """

    def __repr__(self):
        return dict.__repr__(self)


class CoreHeaders(Headers):
    """
    core headers class.
    """

    def to_dict(self):
        """
        gets a dict version of current `Headers` object.

        each duplicate key available in this object, will have
        a list of all values in result dict.

        :rtype: dict
        """

        keys = set(self.keys())
        result = dict()
        for item in keys:
            value = self.getlist(item)
            if len(value) <= 0:
                continue

            if len(value) == 1:
                result[item] = value[0]
            else:
                result[item] = value

        return result



# -*- coding: utf-8 -*-
"""
utils headers module.
"""

from pyrin.core.structs import CoreHeaders


def convert_headers(headers, ignore_null=True):
    """
    converts given headers into a `CoreHeaders` objects.

    :param dict | Headers headers: headers to be converted.

    :param bool ignore_null: specifies that it must return a None
                             value if given headers input is None.
                             otherwise it returns an empty `CoreHeaders`.
                             defaults to True if not provided.


    :rtype: CoreHeaders
    """

    if headers is None and ignore_null is True:
        return headers
    elif headers is None:
        return CoreHeaders()

    return CoreHeaders(headers)



# -*- coding: utf-8 -*-
"""
packaging manager module.
"""

import os
import inspect

from threading import Lock
from importlib import import_module
from time import time

import pyrin.application.services as application_services
import pyrin.configuration.services as config_services
import pyrin.utils.configuration as config_utils
import pyrin.utils.path as path_utils
import pyrin.utils.environment as env_utils
import pyrin.utils.misc as misc_utils

from pyrin.core.mixin import HookMixin
from pyrin.packaging import PackagingPackage
from pyrin.core.structs import DTO, Manager
from pyrin.packaging.base import Package
from pyrin.packaging.enumerations import PackageScopeEnum
from pyrin.packaging.hooks import PackagingHookBase
from pyrin.utils.custom_print import print_info, print_default
from pyrin.packaging.exceptions import InvalidPackageNameError, \
    ComponentModuleNotFoundError, BothUnitAndIntegrationTestsCouldNotBeLoadedError, \
    InvalidPackagingHookTypeError, CircularDependencyDetectedError, PackageNotExistedError, \
    PackageIsIgnoredError, PackageIsDisabledError, SelfDependencyDetectedError, \
    SubPackageDependencyDetectedError, PackageExternalDependencyError


class PackagingManager(Manager, HookMixin):
    """
    packaging manager class.
    """

    _lock = Lock()
    hook_type = PackagingHookBase
    invalid_hook_type_error = InvalidPackagingHookTypeError
    package_class = PackagingPackage
    REQUIRED_PACKAGES = ('application', 'packaging')

    def __init__(self):
        """
        creates a new instance of PackagingManager.
        """

        super().__init__()

        # this flag indicates that application has been loaded.
        # it is required for environments in which server starts
        # multiple threads before application gets loaded.
        self._is_loaded = False

        self._pyrin_package_name = None
        self._required_packages = []

        # holds the names of all application packages that should be loaded.
        self._all_packages = []

        # holds the name of loaded packages.
        self._loaded_packages = []

        # holds the name of disabled packages.
        self._disabled_packages = []

        # a dict containing each package name and all of its dependency package names.
        # in the form of:
        # {str package_name: list[str dependency_package_name]}
        self._dependency_map = DTO()

        # holds the full path of directories that are not a package (not having __init__.py)
        self._not_packages = []

        # configs will be filled from packaging config file.
        self._configs = DTO()

        # holds the root package names in which all test packages are resided.
        self._test_roots = []

        # holds the base roots for different test root packages.
        self._test_roots_bases = []

        # these will keep all loaded components for different
        # categories inside them. extended components in each
        # category are those that extending the exact component
        # of their parent.
        # in the form of: dict[str package_name: list[str] modules]
        self._pyrin_components = DTO()
        self._application_components = DTO()
        self._custom_components = DTO()
        self._test_components = DTO()
        self._unit_test_components = DTO()
        self._integration_test_components = DTO()
        self._extended_application_components = DTO()
        self._other_application_components = DTO()
        self._extended_unit_test_components = DTO()
        self._other_unit_test_components = DTO()
        self._extended_integration_test_components = DTO()
        self._other_integration_test_components = DTO()

    def _create_config_file(self):
        """
        creates packaging config file in application settings path if not available.
        """

        config_services.create_config_file(self.package_class.CONFIG_STORE_NAMES[0],
                                           ignore_on_existed=True)

    def _load_configs(self):
        """
        loads packaging configs from application's settings directory.
        """

        self._configs.clear()
        configs = config_utils.load(self._get_config_file_path())
        self._configs = configs.get('general')
        self._extract_test_roots()

    def _get_config_file_path(self):
        """
        gets packaging config file path.

        it looks for file in top level application settings, but
        if not found it, it uses the file from default settings.
        we have to re-implement this here, because configuration
        services is not present yet to be used.

        :rtype: str
        """

        app_settings_directory = application_services.get_settings_path()
        pyrin_settings_directory = application_services.get_default_settings_path()
        config_file_name = '{store}.ini'.format(store=self.package_class.
                                                CONFIG_STORE_NAMES[0])
        config_path = path_utils.get_first_available_file(app_settings_directory,
                                                          pyrin_settings_directory,
                                                          file_name=config_file_name)

        return config_path

    def _initialize(self):
        """
        initializes required data.
        """

        self._disabled_packages.clear()
        self._not_packages.clear()
        self._dependency_map.clear()
        self._all_packages.clear()
        self._loaded_packages.clear()
        self._required_packages.clear()
        self._pyrin_components.clear()
        self._application_components.clear()
        self._custom_components.clear()
        self._test_components.clear()
        self._unit_test_components.clear()
        self._integration_test_components.clear()
        self._extended_application_components.clear()
        self._other_application_components.clear()
        self._extended_unit_test_components.clear()
        self._other_unit_test_components.clear()
        self._extended_integration_test_components.clear()
        self._other_integration_test_components.clear()

        self._pyrin_package_name = path_utils.get_pyrin_main_package_name()
        self._load_required_packages(self._pyrin_package_name)
        self._load_configs()
        self._resolve_python_path()

    def _load_required_packages(self, pyrin_package):
        """
        loads all required package names.

        these packages are always loaded before any other package and
        they do not need to be handled by packaging package itself.

        :param str pyrin_package: the name of pyrin package.
                                  it would always be `pyrin` in normal cases.
        """

        for item in self.REQUIRED_PACKAGES:
            full_name = '{pyrin}.{package}'.format(pyrin=pyrin_package,
                                                   package=item)
            self._required_packages.append(full_name)
            self._loaded_packages.append(full_name)
            self._all_packages.append(full_name)

    def _extract_test_roots(self):
        """
        extracts the root package names in which all test packages are resided.
        """

        self._test_roots.clear()
        self._test_roots_bases.clear()

        unit = self._configs.unit_test_package
        integration = self._configs.integration_test_package

        if unit not in (None, '') and not unit.isspace():
            unit_root = unit.split('.')
            if len(unit_root) > 1:
                unit_root.pop()

            self._test_roots.append('.'.join(unit_root))
            self._test_roots_bases.append(unit_root[0])

        if integration not in (None, '') and not integration.isspace():
            integration_root = integration.split('.')
            if len(integration_root) > 1:
                integration_root.pop()

            value = '.'.join(integration_root)
            if value not in self._test_roots:
                self._test_roots.append(value)

            if integration_root[0] not in self._test_roots_bases:
                self._test_roots_bases.append(integration_root[0])

    def load_components(self, **options):
        """
        loads required packages and modules for application startup.

        :raises BothUnitAndIntegrationTestsCouldNotBeLoadedError: both unit and integration
                                                                  tests could not be loaded
                                                                  error.
        :raises PackageIsIgnoredError: package is ignored error.
        :raises PackageIsDisabledError: package is disabled error.
        :raises PackageNotExistedError: package not existed error.
        :raises SelfDependencyDetectedError: self dependency detected error.
        :raises SubPackageDependencyDetectedError: sub-package dependency detected error.
        :raises CircularDependencyDetectedError: circular dependency detected error.
        :raises PackageExternalDependencyError: package external dependency error.
        """

        if self._is_loaded is True:
            return

        with self._lock:
            if self._is_loaded is True:
                return

            start_time = time()
            self._initialize()

            print_info('Loading application components...')

            self._find_pyrin_loadable_components()
            self._find_other_loadable_components()

            self._load_components(self._pyrin_components, **options)
            self._load_components(self._extended_application_components, **options)
            self._load_components(self._other_application_components, **options)
            self._load_components(self._custom_components, **options)
            self._load_tests(**options)

            self._after_packages_loaded()

            print_info('Total of [{count}] packages loaded.'
                       .format(count=len(self._loaded_packages)))

            self._create_config_file()
            self._is_loaded = True
            end_time = time()
            duration = '{:0.1f}'.format((end_time - start_time) * 1000)
            print_info('Application loaded in [{duration}] milliseconds.'
                       .format(duration=duration))

            pyrin_version = application_services.get_pyrin_version()
            print_info('Pyrin version: [{version}].'.format(version=pyrin_version))

    def _load_tests(self, **options):
        """
        loads test packages if needed.

        :raises BothUnitAndIntegrationTestsCouldNotBeLoadedError: both unit and integration
                                                                  tests could not be loaded
                                                                  error.
        :raises PackageIsIgnoredError: package is ignored error.
        :raises PackageIsDisabledError: package is disabled error.
        :raises PackageNotExistedError: package not existed error.
        :raises SelfDependencyDetectedError: self dependency detected error.
        :raises SubPackageDependencyDetectedError: sub-package dependency detected error.
        :raises CircularDependencyDetectedError: circular dependency detected error.
        :raises PackageExternalDependencyError: package external dependency error.
        """

        if self._configs.load_unit_test is True and \
                self._configs.load_integration_test is True:
            raise BothUnitAndIntegrationTestsCouldNotBeLoadedError('Both unit and '
                                                                   'integration tests '
                                                                   'could not be loaded '
                                                                   'at the same time.')

        if self._configs.load_unit_test is True or \
                self._configs.load_integration_test is True:
            self._load_components(self._test_components, **options)

            if self._configs.load_unit_test is True:
                self._load_components(self._extended_unit_test_components, **options)
                self._load_components(self._other_unit_test_components, **options)

            elif self._configs.load_integration_test is True:
                self._load_components(self._extended_integration_test_components, **options)
                self._load_components(self._other_integration_test_components, **options)

    def _after_packages_loaded(self):
        """
        this method will call `after_packages_loaded` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_packages_loaded()

    def _package_loaded(self, package_name, **options):
        """
        this method will call `package_loaded` method of all registered hooks.

        :param str package_name: name of the loaded package.
        """

        for hook in self._get_hooks():
            hook.package_loaded(package_name, **options)

    def load(self, module_name, **options):
        """
        loads the specified module.

        :param str module_name: full module name.
                                example module_name = `pyrin.application.decorators`.

        :rtype: Module
        """

        module = import_module(module_name)
        return module

    def _load_component(self, package_name, module_names, component_name, **options):
        """
        loads the given component.

        :param str package_name: full package name to be loaded.
        :param list[str] module_names: full module names to be loaded.
        :param str component_name: component name of this package.

        :raises ComponentModuleNotFoundError: component module not found error.
        """

        self.load(package_name)

        # component module should be loaded first if available, in case of
        # any other module needed package services in top level objects.
        component_module = None
        if component_name is not None:
            component_module = self._merge_module_name(package_name, component_name)

        if component_module is not None and component_module in module_names:
            self.load(component_module, **options)
        elif component_module is not None and component_module not in module_names:
            raise ComponentModuleNotFoundError('Component module [{name}] not '
                                               'found in [{package}] package.'
                                               .format(name=component_module,
                                                       package=package_name))

        for module in module_names:
            if module != component_module:
                self.load(module, **options)

        self._loaded_packages.append(package_name)
        self._package_loaded(package_name, **options)

        print_default('[{package}] package loaded. including [{module_count}] modules.'
                      .format(package=package_name,
                              module_count=len(module_names)))

        if package_name == self._pyrin_package_name:
            for item in self._required_packages:
                print_default('[{package}] package loaded.'
                              .format(package=item))

    def _load_components(self, components, **options):
        """
        loads the given components considering their dependency on each other.

        :param dict components: full package names and their
                                modules to be loaded.

        :note components: dict[str package_name: list[str] modules]

        :raises PackageIsIgnoredError: package is ignored error.
        :raises PackageIsDisabledError: package is disabled error.
        :raises PackageNotExistedError: package not existed error.
        :raises SelfDependencyDetectedError: self dependency detected error.
        :raises SubPackageDependencyDetectedError: sub-package dependency detected error.
        :raises CircularDependencyDetectedError: circular dependency detected error.
        :raises PackageExternalDependencyError: package external dependency error.
        """

        # a dictionary containing all dependent package names and their respective modules.
        # in the form of {str package_name: [str module]}.
        dependent_components = DTO()

        for package in components:
            dependencies = []
            package_class = self._get_package_class(package)
            if package_class is not None:
                dependencies = package_class.DEPENDS

            self._validate_dependencies(package, dependencies)

            # checking whether this package has any dependencies.
            # if so, check those dependencies have been loaded or not.
            # if not, then put this package into dependent_packages and
            # load it later. otherwise load it now.
            if (len(dependencies) <= 0 or
                self._is_dependencies_loaded(dependencies) is True) and \
                    self._is_parent_loaded(package) is True:

                instance = None
                if package_class is not None:
                    instance = package_class()
                    instance.load_configs(config_services)

                component_name = None
                if instance is not None:
                    component_name = instance.COMPONENT_NAME
                self._load_component(package, components[package], component_name, **options)
            else:
                dependent_components[package] = components[package]

        # now, go through dependent components if any, and try to load them.
        if len(dependent_components) > 0:
            self._load_components(dependent_components, **options)

    def _validate_dependencies(self, package_name, dependencies):
        """
        validates that given package's dependencies have no problem.

        it checks for different problems such as self and circular
        dependencies, unavailable dependencies, external dependencies and more.

        it raises an error if a problem has been detected.
        for example, if `pyrin.database` has a dependency on `pyrin.logging`
        and `pyrin.logging` also has a dependency on `pyrin.database`, this
        method raises an error.

        :param str package_name: package name.
        :param list[str] dependencies: list of given package's dependencies.

        :raises PackageIsIgnoredError: package is ignored error.
        :raises PackageIsDisabledError: package is disabled error.
        :raises PackageNotExistedError: package not existed error.
        :raises SelfDependencyDetectedError: self dependency detected error.
        :raises SubPackageDependencyDetectedError: sub-package dependency detected error.
        :raises CircularDependencyDetectedError: circular dependency detected error.
        :raises PackageExternalDependencyError: package external dependency error.
        """

        dependencies = misc_utils.make_iterable(dependencies, list)
        self._check_dependencies_exist(package_name, dependencies)

        self._dependency_map[package_name] = dependencies
        if len(dependencies) <= 0:
            return

        if package_name in dependencies:
            raise SelfDependencyDetectedError('Package [{source}] has a dependency on itself. '
                                              'it is a mistake to depend a package on itself.'
                                              .format(source=package_name))

        for item in dependencies:
            if self._contains(package_name, item) is True:
                raise SubPackageDependencyDetectedError('Package [{root}] has a dependency on '
                                                        'its sub-package [{child}]. it is a '
                                                        'mistake to depend a package on its '
                                                        'own sub-packages.'
                                                        .format(root=package_name, child=item))

            reverse_dependencies = self._dependency_map.get(item)
            if reverse_dependencies is not None:
                if package_name in reverse_dependencies:
                    raise CircularDependencyDetectedError('There is a circular dependency '
                                                          'between [{source}] and [{reverse}] '
                                                          'packages.'
                                                          .format(source=package_name,
                                                                  reverse=item))

            if self._is_valid_external_dependency(package_name, item) is False:
                raise PackageExternalDependencyError('Package [{source}] has a dependency '
                                                     'on package [{other}] which is from '
                                                     'an outer scope. a package could not be '
                                                     'dependent on outer scope packages.'
                                                     .format(source=package_name,
                                                             other=item))

    def _check_dependencies_exist(self, package_name, dependencies):
        """
        checks that given dependency packages are available in the application scope.

        :param str package_name: package name.
        :param list[str] dependencies: list of given package's dependencies.

        :raises PackageIsIgnoredError: package is ignored error.
        :raises PackageIsDisabledError: package is disabled error.
        :raises PackageNotExistedError: package not existed error.
        """

        if dependencies is None:
            return

        for item in dependencies:
            if item not in self._all_packages:
                base_message = 'Provided dependency package [{name}] ' \
                               'specified in [{source}] package,' \
                    .format(name=item, source=package_name)

                if self._is_ignored_package(item):
                    raise PackageIsIgnoredError('{base_message} is ignored in '
                                                'packaging config store.'
                                                .format(base_message=base_message))
                if self._is_disabled_package(item):
                    raise PackageIsDisabledError('{base_message} is disabled.'
                                                 .format(base_message=base_message))

                raise PackageNotExistedError('{base_message} does not exist.'
                                             .format(base_message=base_message))

    def _is_valid_external_dependency(self, package_name, dependency):
        """
        gets a value indicating that given dependency package is a valid external dependency.

        a package could not be dependent on external scope packages.
        for example `pyrin.database` package could not be dependent on
        `my_app.database` package which is from outer scope.

        note that a package could be dependent on another package from inner
        scopes, although it is nonsense. so a package like `my_app.database`
        could be dependent on `pyrin.database` but it is not required to put this
        dependency in package info because the `pyrin.database` package is always
        gets loaded before any package of outer scope application gets loaded.

        :param str package_name: package name to check its dependency.
        :param str dependency: dependency package name to be checked.

        :rtype: bool
        """

        package_scope = self._get_package_scope(package_name)
        dependency_scope = self._get_package_scope(dependency)

        if package_scope == PackageScopeEnum.UNKNOWN or \
                dependency_scope == PackageScopeEnum.UNKNOWN:
            return False

        if (self._is_unit_test_scope(package_scope) and
            self._is_integration_test_scope(dependency_scope)) or \
                (self._is_integration_test_scope(package_scope) and
                 self._is_unit_test_scope(dependency_scope)):
            return False

        return dependency_scope <= package_scope

    def _is_integration_test_scope(self, scope):
        """
        gets a value indicating that given value belongs to any of integration test scopes.

        :param int scope: the scope to be checked.
        :enum scope:
            PYRIN = 0
            EXTENDED_APPLICATION = 1
            OTHER_APPLICATION = 2
            CUSTOM_APPLICATION = 3
            TEST = 4
            EXTENDED_UNIT_TEST = 5
            OTHER_UNIT_TEST = 6
            EXTENDED_INTEGRATION_TEST = 7
            OTHER_INTEGRATION_TEST = 8
            UNKNOWN = 100

        :rtype: bool
        """

        return scope in (PackageScopeEnum.OTHER_INTEGRATION_TEST,
                         PackageScopeEnum.EXTENDED_INTEGRATION_TEST)

    def _is_unit_test_scope(self, scope):
        """
        gets a value indicating that given value belongs to any of unit test scopes.

        :param int scope: the scope to be checked.
        :enum scope:
            PYRIN = 0
            EXTENDED_APPLICATION = 1
            OTHER_APPLICATION = 2
            CUSTOM_APPLICATION = 3
            TEST = 4
            EXTENDED_UNIT_TEST = 5
            OTHER_UNIT_TEST = 6
            EXTENDED_INTEGRATION_TEST = 7
            OTHER_INTEGRATION_TEST = 8
            UNKNOWN = 100

        :rtype: bool
        """

        return scope in (PackageScopeEnum.OTHER_UNIT_TEST,
                         PackageScopeEnum.EXTENDED_UNIT_TEST)

    def _get_package_scope(self, package_name):
        """
        gets the scope of given package.

        :param str package_name: package name to get its scope.

        :returns: the scope of given package.
        :enum scope:
            PYRIN = 0
            EXTENDED_APPLICATION = 1
            OTHER_APPLICATION = 2
            CUSTOM_APPLICATION = 3
            TEST = 4
            EXTENDED_UNIT_TEST = 5
            OTHER_UNIT_TEST = 6
            EXTENDED_INTEGRATION_TEST = 7
            OTHER_INTEGRATION_TEST = 8
            UNKNOWN = 100

        :rtype: int
        """

        if self._is_pyrin_package(package_name) is True:
            return PackageScopeEnum.PYRIN

        if self._is_extended_application_package(package_name) is True:
            return PackageScopeEnum.EXTENDED_APPLICATION

        if self._is_other_application_package(package_name) is True:
            return PackageScopeEnum.OTHER_APPLICATION

        if self._is_custom_package(package_name) is True:
            return PackageScopeEnum.CUSTOM_APPLICATION

        if self._is_test_package(package_name) is True:
            return PackageScopeEnum.TEST

        if self._is_extended_unit_test_package(package_name) is True:
            return PackageScopeEnum.EXTENDED_UNIT_TEST

        if self._is_other_unit_test_package(package_name) is True:
            return PackageScopeEnum.OTHER_UNIT_TEST

        if self._is_extended_integration_test_package(package_name) is True:
            return PackageScopeEnum.EXTENDED_INTEGRATION_TEST

        if self._is_other_integration_test_package(package_name) is True:
            return PackageScopeEnum.OTHER_INTEGRATION_TEST

        return PackageScopeEnum.UNKNOWN

    def _find_pyrin_loadable_components(self):
        """
        finds all package and module names that should be loaded from pyrin package.
        """

        pyrin_root_path = application_services.get_pyrin_root_path()
        pyrin_path = application_services.get_pyrin_main_package_path()
        self._find_loadable_components(pyrin_root_path, include=pyrin_path)

    def _find_other_loadable_components(self):
        """
        finds all package and module names that should be loaded from other packages.

        for example application and tests.
        """

        application_root_path = application_services.get_application_root_path()
        working_directory = self.get_working_directory(application_root_path)
        pyrin_path = application_services.get_pyrin_main_package_path()
        self._find_loadable_components(working_directory, exclude=pyrin_path)

    def get_working_directory(self, root_path):
        """
        gets working directory path according to given root path.

        working directory is where the root application and test package are resided.
        this is required when application starts from any of test applications.
        then we should move root path up, to the correct root to be able to
        include real application packages too.
        if the application has been started from real application, this method
        returns the same input.

        :param str root_path: root path to get working directory from.

        :rtype: str
        """

        roots = tuple(item.replace('.', os.path.sep) for item in self._test_roots)
        if len(roots) > 0:
            for item in roots:
                if root_path.endswith(item):
                    return root_path.replace(item, '').rstrip(os.path.sep)

        return root_path

    def _resolve_python_path(self):
        """
        resolves python path to put in `PYTHONPATH` variable.
        """

        application_root_path = application_services.get_application_root_path()
        env_utils.set_python_path(self.get_working_directory(application_root_path))

    def _find_loadable_components(self, root_path, include=None,
                                  exclude=None, **options):
        """
        finds all package and module names that should be loaded included in given root path.

        :param str root_path: root path to look for components inside it.

        :param str | list[str] include: specify full directory names inside the
                                        root path to just loop inside those.
                                        otherwise it loops in all available
                                        directories.

        :param str | list[str] exclude: specify full directory names inside the
                                        root path to ignore them. otherwise
                                        it loops in all available directories.
        """

        include = misc_utils.make_iterable(include, list)
        exclude = misc_utils.make_iterable(exclude, list)

        for root, directories, file_names in os.walk(root_path, followlinks=True):
            temp_dirs = list(directories)
            for single_dir in temp_dirs:
                visiting_path = os.path.abspath(os.path.join(root, single_dir))
                if self._should_visit(include, exclude, visiting_path) is False:
                    directories.remove(single_dir)

            for directory in directories:
                combined_path = os.path.join(root, directory)

                if not self._is_package(combined_path):
                    continue

                package_name = self._get_package_name(combined_path, root_path)
                if self._is_ignored_package(package_name):
                    continue

                package_class = self._get_package_class(package_name)
                if package_class is not None and package_class.ENABLED is False:
                    self._disabled_packages.append(package_name)
                    continue

                if self._is_disabled_package(package_name):
                    continue

                self._all_packages.append(package_name)

                if self._is_pyrin_package(package_name):
                    self._pyrin_components[package_name] = []
                elif self._is_custom_package(package_name):
                    self._custom_components[package_name] = []
                elif self._is_unit_test_package(package_name):
                    self._unit_test_components[package_name] = []
                elif self._is_integration_test_package(package_name):
                    self._integration_test_components[package_name] = []
                elif self._is_test_package(package_name):
                    self._test_components[package_name] = []
                else:
                    self._application_components[package_name] = []

                files = os.listdir(combined_path)
                for file_name in files:
                    if not self._is_module(file_name):
                        continue

                    module_name = file_name.replace('.py', '')
                    full_module_name = self._get_module_name(package_name, module_name)
                    if self._is_ignored_module(full_module_name):
                        continue

                    if self._is_pyrin_module(full_module_name):
                        self._pyrin_components[package_name].append(full_module_name)
                    elif self._is_custom_module(full_module_name):
                        self._custom_components[package_name].append(full_module_name)
                    elif self._is_unit_test_module(full_module_name):
                        self._unit_test_components[package_name].append(full_module_name)
                    elif self._is_integration_test_module(full_module_name):
                        self._integration_test_components[package_name].append(full_module_name)
                    elif self._is_test_module(full_module_name):
                        self._test_components[package_name].append(full_module_name)
                    else:
                        self._application_components[package_name].append(full_module_name)

        self._detach_all()

    def _detach_all(self):
        """
        detaches all founded components into extended and other components.
        """

        self._extended_application_components, \
            self._other_application_components = self._detach_extended_packages(
                list(self._pyrin_components.keys()), self._application_components)

        unit_test_base_components = self._application_components
        integration_test_base_components = self._application_components
        if len(self._application_components) <= 0:
            unit_test_base_components = self._pyrin_components
            integration_test_base_components = self._pyrin_components

        self._extended_unit_test_components, \
            self._other_unit_test_components = self._detach_extended_packages(
                list(unit_test_base_components.keys()),
                self._unit_test_components,
                main_package_name=self._configs.unit_test_package)

        self._extended_integration_test_components, \
            self._other_integration_test_components = self._detach_extended_packages(
                list(integration_test_base_components.keys()),
                self._integration_test_components,
                main_package_name=self._configs.integration_test_package)

    def _detach_extended_packages(self, base_components, components, **options):
        """
        detaches components which extend existing base components from new components.

        :param list[str] base_components: base component names.

        :param dict components: components which some of
                                them extend base components.

        :note components: dict[str package_name: list[str] modules]

        :keyword str main_package_name: main package name of given components.
                                        if not provided, it will be assumed equal
                                        to the first part of component name.
                                        for example: if component is `pyrin.api.schema`
                                        then, it will be assumed equal to `pyrin`.
                                        if the package structure has a root with more
                                        than one part, it should be provided manually.

        :returns: tuple[dict extended_components, dict other_components]

        :note extended_components: dict[str package_name: list[str] modules]
        :note other_components: dict[str package_name: list[str] modules]

        :rtype: tuple[dict, dict]
        """

        extended_components = DTO()
        other_components = DTO()
        if len(components) > 0:
            component_keys = list(components.keys())
            root_name = options.get('main_package_name', None)
            if root_name is None:
                root_name = self._get_root_package(component_keys[0])
            base_names = [self._replace_root_package(item, root_name) for item in base_components]
            for package in components:
                if package in base_names:
                    extended_components[package] = components[package]
                else:
                    other_components[package] = components[package]

        return extended_components, other_components

    def _replace_root_package(self, old_package, root_name):
        """
        replaces root package name in old package with given root name.

        :param str old_package: old package fully qualified name.
                                in the form of: `pyrin.api`

        :param str root_name: root name to be put in old package name.
                              in the form of: `my_root`
        :rtype: str
        """

        parts = old_package.split('.')
        parts.pop(0)
        parts.insert(0, root_name)

        return '.'.join(parts)

    def _get_root_package(self, component_name):
        """
        gets the root package of given component.

        :param str component_name: full package or module name.

        :rtype: str
        """

        return path_utils.get_main_package_name(component_name)

    def _is_included(self, include, visiting_path):
        """
        returns a value indicating that the given visiting path is under include path.

        :param list[str] include: full directory names inside the
                                  root path to just loop inside those.
                                  otherwise it loops in all available
                                  directories.

        :param str visiting_path: full path which must be checked for inclusion.

        :rtype: bool
        """

        if include is None or len(include) <= 0:
            return True

        for path in include:
            if visiting_path.startswith(path):
                return True

        return False

    def _is_excluded(self, exclude, visiting_path):
        """
        returns a value indicating that the given visiting path is under exclude path.

        :param list[str] exclude: full directory names inside the
                                  root path to ignore them. otherwise
                                  it loops in all available directories.

        :param str visiting_path: full path which must be checked for exclusion.

        :rtype: bool
        """

        if exclude is None or len(exclude) <= 0:
            return False

        for path in exclude:
            if visiting_path.startswith(path):
                return True

        return False

    def _should_visit(self, include, exclude, visiting_path):
        """
        gets a value indicating that given path should be visited.

        :param list[str] include: full directory names inside the
                                  root path to just loop inside those.
                                  otherwise it loops in all available
                                  directories.

        :param list[str] exclude: full directory names inside the
                                  root path to ignore them. otherwise
                                  it loops in all available directories.

        :param str visiting_path: full path which must be checked for exclusion.

        :rtype: bool
        """

        return self._is_included(include, visiting_path) is True and \
            self._is_excluded(exclude, visiting_path) is False

    def _is_equal(self, condition, full_module_or_package):
        """
        gets a value indicating that given name is equal with given condition.

        :param str condition: condition to compare with name.
                              for example: `*.database`, then all values
                              that have database in their second part
                              and their parts count are equal to or greater
                              than condition parts count will be recognized as equal.

        :param str full_module_or_package: module or package name to be compared.
                                           for example: `pyrin.database`.

        :rtype: bool
        """

        condition_parts = condition.split('.')
        full_module_or_package_parts = full_module_or_package.split('.')

        if len(condition_parts) > len(full_module_or_package_parts):
            return False

        for index, item in enumerate(condition_parts):
            if item != '*' and item != full_module_or_package_parts[index]:
                return False

        return True

    def _is_disabled_package(self, package_name):
        """
        gets a value indicating that given package should be considered as disabled.

        it will be detected based on parent packages of this package.

        :param str package_name: full package name.
                                 example package_name = `pyrin.database`.

        :rtype: bool
        """

        for disabled in self._disabled_packages:
            if package_name.startswith(disabled) or self._is_equal(disabled, package_name):
                return True

        return False

    def _is_ignored_package(self, package_name):
        """
        gets a value indicating that given package should be ignored.

        :param str package_name: full package name.
                                 example package_name = `pyrin.database`.

        :rtype: bool
        """

        if package_name in self._required_packages:
            return True

        for ignored in self._configs.ignored_packages:
            if package_name.startswith(ignored) or self._is_equal(ignored, package_name):
                return True

        return False

    def _is_ignored_module(self, module_name):
        """
        gets a value indicating that given module should be ignored.

        :param str module_name: full module name.
                                example module_name = `pyrin.api.error_handlers`.

        :rtype: bool
        """

        for ignored in self._configs.ignored_modules:
            if module_name.endswith(ignored) or self._is_equal(ignored, module_name):
                return True

        return False

    def _is_pyrin_component(self, component_name):
        """
        gets a value indicating that given component is a pyrin component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        return self._contains(self._pyrin_package_name, component_name)

    def _is_pyrin_package(self, package_name):
        """
        gets a value indicating that given package is a pyrin package.

        :param str package_name: full package name.
                                 example package_name = 'pyrin.api'

        :rtype: bool
        """

        return self._is_pyrin_component(package_name)

    def _is_pyrin_module(self, module_name):
        """
        gets a value indicating that given module is a pyrin module.

        :param str module_name: full module name.
                                example module_name = 'pyrin.api.error_handlers'

        :rtype: bool
        """

        return self._is_pyrin_component(module_name)

    def _is_other_application_package(self, package_name):
        """
        gets a value indicating that given package is from other application packages.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._other_application_components

    def _is_extended_application_package(self, package_name):
        """
        gets a value indicating that given package is an extended application package.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._extended_application_components

    def _is_custom_component(self, component_name):
        """
        gets a value indicating that given component is a custom component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        for custom in self._configs.custom_packages:
            if self._contains(custom, component_name) is True:
                return True

        return False

    def _is_custom_package(self, package_name):
        """
        gets a value indicating that given package is a custom package.

        :param str package_name: full package name.
                                 example package_name = 'custom.api'

        :rtype: bool
        """

        return self._is_custom_component(package_name)

    def _is_custom_module(self, module_name):
        """
        gets a value indicating that given module is a custom module.

        :param str module_name: full module name.
                                example module_name = 'custom.api.error_handlers'

        :rtype: bool
        """

        return self._is_custom_component(module_name)

    def _is_unit_test_component(self, component_name):
        """
        gets a value indicating that given component is a unit test component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        return self._contains(self._configs.unit_test_package, component_name)

    def _is_unit_test_package(self, package_name):
        """
        gets a value indicating that given package is a unit test package.

        :param str package_name: full package name.
                                 example package_name = 'test.api'

        :rtype: bool
        """

        return self._is_unit_test_component(package_name)

    def _is_unit_test_module(self, module_name):
        """
        gets a value indicating that given module is a unit test module.

        :param str module_name: full module name.
                                example module_name = 'test.api.error_handlers'

        :rtype: bool
        """

        return self._is_unit_test_component(module_name)

    def _is_other_unit_test_package(self, package_name):
        """
        gets a value indicating that given package is from other unit packages.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._other_unit_test_components

    def _is_extended_unit_test_package(self, package_name):
        """
        gets a value indicating that given package is an extended unit test package.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._extended_unit_test_components

    def _is_integration_test_component(self, component_name):
        """
        gets a value indicating that given component is an integration test component.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        return self._contains(self._configs.integration_test_package, component_name)

    def _is_integration_test_package(self, package_name):
        """
        gets a value indicating that given package is an integration test package.

        :param str package_name: full package name.
                                 example package_name = 'test.api'

        :rtype: bool
        """

        return self._is_integration_test_component(package_name)

    def _is_integration_test_module(self, module_name):
        """
        gets a value indicating that given module is an integration test module.

        :param str module_name: full module name.
                                example module_name = 'test.api.error_handlers'

        :rtype: bool
        """

        return self._is_integration_test_component(module_name)

    def _is_other_integration_test_package(self, package_name):
        """
        gets a value indicating that given package is from other integration test packages.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._other_integration_test_components

    def _is_extended_integration_test_package(self, package_name):
        """
        gets a value indicating that given package is an extended integration test package.

        :param str package_name: full package name.

        :rtype: bool
        """

        return package_name in self._extended_integration_test_components

    def _is_test_component(self, component_name):
        """
        gets a value indicating that given component is a test component.

        this method specifies those components that are test components but
        won't be included neither in unit nor integration test components.

        :param str component_name: full package or module name.

        :rtype: bool
        """

        if not self._is_integration_test_component(component_name) and \
                not self._is_unit_test_component(component_name):
            for root in self._test_roots_bases:
                is_test = self._contains(root, component_name)
                if is_test is True:
                    return True

        return False

    def _is_test_package(self, package_name):
        """
        gets a value indicating that given package is a test package.

        this method specifies those packages that are test packages but
        won't be included neither in unit nor integration test packages.

        :param str package_name: full package name.
                                 example package_name = 'test.api'

        :rtype: bool
        """

        return self._is_test_component(package_name)

    def _is_test_module(self, module_name):
        """
        gets a value indicating that given module is a test module.

        this method specifies those modules that are test modules but
        won't be included neither in unit nor integration test modules.

        :param str module_name: full module name.
                                example module_name = 'test.api.error_handlers'

        :rtype: bool
        """

        return self._is_test_component(module_name)

    def _contains(self, root, component_name):
        """
        gets a value indicating that given component
        qualified name, is belonging to the provided root.

        :param str root: root name that should be checked for component existence.
                         example root = `application.custom`

        :param str component_name: component name that should be
                                   checked for existence in root.
                                   example component_name = `application.custom.api`

        :rtype: bool
        """

        parts_component = component_name.split('.')
        parts_root = root.split('.')
        if len(parts_component) < len(parts_root):
            return False

        component_root = '.'.join(parts_component[0:len(parts_root)])

        return component_root == root

    def _get_package_name(self, path, root_path):
        """
        gets the full package name for provided path.

        :param str path: full path of package.
                         example path = `/home/src/pyrin/database`.

        :param str root_path: root path in which this path is located.
                              example root_path = `/home/src`

        :rtype: str
        """

        return path_utils.get_package_name(path, root_path)

    def _merge_module_name(self, package_name, component_name):
        """
        merges package and component name and gets the fully qualified module name.

        :param str package_name: package name.
                                 example package_name = `pyrin.database`.

        :param str component_name: component name.
                                   example component_name = `database.component`.

        :rtype: str
        """

        parts = component_name.split('.')
        return self._get_module_name(package_name, parts[-1])

    def _get_module_name(self, package_name, module_name):
        """
        gets the full module name.

        :param str package_name: package name.
                                 example package_name = `pyrin.database`.

        :param str module_name: module name.
                                example module_name = `api`.

        :rtype: str
        """

        return '{package}.{module}'.format(package=package_name, module=module_name)

    def _is_package(self, path):
        """
        gets a value indicating that given path belongs to a python package.

        it simply checks that `__init__` module exists or not.

        :param str path: full path of package.
                         example path = `/home/src/pyrin/database`.

        :rtype: bool
        """

        is_package = self._has_module(path, '__init__')
        if is_package is False:
            self._not_packages.append(path)

        return is_package and self._parent_is_package(path)

    def _parent_is_package(self, path):
        """
        gets a value indicating that the parent of given path is a python package.

        :param str path: full path of package.
                         example path = `/home/src/pyrin/database`.

        :rtype: bool
        """

        for not_package in self._not_packages:
            if path.startswith(not_package):
                return False

        return True

    def _is_module(self, file_name):
        """
        gets a value indicating that given file is a standalone python module.

        excluding `__init__` module which belongs to package.
        it simply checks that file name ends with '.py' and not being `__init__.py`.

        :param str file_name: file name.
                              example file_name = `services.py`
        :rtype: bool
        """

        return file_name.endswith('.py') and '__init__.py' not in file_name

    def _has_module(self, path, module_name):
        """
        gets a value indicating that given module exists in specified path.

        :param str path: path to check module availability in it.
                         example path = `/home/src/pyrin/database`.

        :param str module_name: module name.
                                example module_name = `__init__`.

        :rtype: bool
        """

        return os.path.isfile(os.path.join(path, '{module}.py'.format(module=module_name)))

    def _get_package_class(self, package_name):
        """
        gets the package class implemented in given package if available, otherwise returns None.

        :param str package_name: full package name.
                                 example package_name = `pyrin.api`.

        :rtype: type
        """

        module = self.load(package_name)
        package_class = None

        for cls in module.__dict__.values():
            if inspect.isclass(cls) and cls is not Package and issubclass(cls, Package):
                package_class = cls
        return package_class

    def _is_dependencies_loaded(self, dependencies):
        """
        gets a value indicating that given dependencies has been already loaded.

        :param list[str] dependencies: full dependency names.
                                       example dependencies = `pyrin.logging`

        :rtype: bool
        """

        for dependency in dependencies:
            if dependency not in self._loaded_packages:
                return False

        return True

    def _is_parent_loaded(self, package_name):
        """
        gets a value indicating that given package's parent package has been loaded.

        :param str package_name: full package name.
                                 example package_name = `pyrin.encryption.handlers`

        :raises InvalidPackageNameError: invalid package name error.

        :rtype: bool
        """

        items = package_name.split('.')
        parent_package = None

        length = len(items)
        if length == 1:
            parent_package = items[0]
        elif length > 1:
            parent_package = '.'.join(items[:-1])
        else:
            raise InvalidPackageNameError('Input package name [{package_name}] is invalid.'
                                          .format(package_name=package_name))

        # application root packages like `pyrin`, has no
        # parent so it should always return `True` for them.
        if parent_package == package_name:
            return True

        return self._is_dependencies_loaded([parent_package])

    def get_loaded_packages(self):
        """
        gets the name of all loaded packages.

        :rtype: list[str]
        """

        return list(self._loaded_packages)

    def is_package_loaded(self, name):
        """
        gets a value indicating that given package is loaded.

        :param str name: package fully qualified name.

        :rtype: bool
        """

        return name in self._loaded_packages



# -*- coding: utf-8 -*-
"""
celery audit services module.
"""

from pyrin.application.services import get_component
from pyrin.task_queues.celery.audit import CeleryAuditPackage


def inspect(**options):
    """
    inspects the status of celery.

    it returns a tuple of two values. first value is a dict containing the inspection
    data. and the second value is a bool value indicating that inspection has been
    succeeded or failed.

    :keyword bool traceback: specifies that on failure report, it must include
                             the traceback of errors.
                             defaults to True if not provided.

    :keyword bool raise_error: specifies that it must raise error
                               if any of registered audits failed
                               instead of returning a failure response.
                               defaults to False if not provided.

    :rtype: tuple[dict, bool]
    """

    return get_component(CeleryAuditPackage.COMPONENT_NAME).inspect(**options)


def perform_job():
    """
    performs a dummy job.
    """

    return get_component(CeleryAuditPackage.COMPONENT_NAME).perform_job()



# -*- coding: utf-8 -*-
"""
response status package.
"""

from pyrin.packaging.base import Package


class ResponseStatusPackage(Package):
    """
    response status package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'processor.response.status.component'



# -*- coding: utf-8 -*-
"""
string normalizer handlers filter module.
"""

import re

import pyrin.utils.string as string_utils

from pyrin.utilities.string.normalizer.decorators import string_normalizer
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.handlers.base import StringNormalizerBase, \
    ReplaceNormalizerBase


@string_normalizer()
class FilterNormalizer(StringNormalizerBase):
    """
    filter normalizer class.

    this normalizer removes given list of values from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of FilterNormalizer.
        """

        super().__init__(NormalizerEnum.FILTER, 430, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :keyword list[str] filters: values to be removed from string.
        :keyword bool ignore_case: filter matches in case-insensitive way.
                                   defaults to True if not provided.

        :returns: normalized value.
        :rtype: str
        """

        filters = options.get('filters')
        if filters is None:
            return value

        flags = re.IGNORECASE
        ignore_case = options.get('ignore_case', True)
        if ignore_case is False:
            flags = 0

        filters = string_utils.sort_by_length(filters, reverse=True)
        for item in filters:
            value = re.sub(item, '', value, flags=flags)

        return value


@string_normalizer()
class PersianSignFilterNormalizer(ReplaceNormalizerBase):
    """
    persian sign filter normalizer class.

    this normalizer removes all common persian signs from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of PersianSignFilterNormalizer.
        """

        replace_map = {'،': None, '؛': None, 'ـ': None,
                       '٬': None, '٪': None, '؟': None}

        super().__init__(NormalizerEnum.PERSIAN_SIGN, 420, replace_map, **options)


@string_normalizer()
class LatinSignFilterNormalizer(ReplaceNormalizerBase):
    """
    latin sign filter normalizer class.

    this normalizer removes all common latin signs from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of LatinSignFilterNormalizer.
        """

        replace_map = {',': None, '.': None, ':': None, ';': None, '-': None,
                       '+': None, '_': None, '`': None, '"': None, "'": None,
                       '@': None, '#': None, '%': None, '^': None, '*': None,
                       '$': None, '(': None, ')': None, '{': None, '}': None,
                       '|': None, '~': None, '?': None, '!': None, '/': None,
                       '\\': None, '<': None, '>': None, '[': None, ']': None}

        super().__init__(NormalizerEnum.LATIN_SIGN, 410, replace_map, **options)



# -*- coding: utf-8 -*-
"""
orm sql schema base module.
"""

from uuid import UUID

import sqlalchemy.dialects.mssql as mssql_types
import sqlalchemy.dialects.postgresql as pg_types
import sqlalchemy.dialects.sybase.base as sybase_types

from sqlalchemy.exc import ArgumentError
from sqlalchemy import Column, util, ARRAY

from pyrin.core.globals import LIST_TYPES
from pyrin.utils.sqlalchemy import check_constraint, range_check_constraint
from pyrin.caching.decorators import cached_property
from pyrin.database.orm.sql.operators.base import CoreColumnOperators
from pyrin.database.orm.sql.schema.exceptions import InvalidCheckConstraintError, \
    CheckConstraintConflictError


class CoreColumn(Column, CoreColumnOperators):
    """
    core column class.

    all application models columns must be an instance of this class.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                               database-side default generation function,
                                               such as a trigger. this indicates to sqlalchemy
                                               that a newly generated value will be available
                                               after updates. this construct does not actually
                                               implement any kind of generation function within
                                               the database, which instead must be specified
                                               separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.

        :keyword object | callable min_value: minimum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword object | callable max_value: maximum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword list | callable check_in: list of valid values for this column.
                                           it could also be a callable without any inputs.
                                           if a non-callable is provided and the column is
                                           not a primary key and also column name is provided,
                                           it will result in check constraint generation on
                                           database. otherwise it will be ignored and could
                                           be used in validators.
                                           defaults to None if not provided.

        :keyword list | callable check_not_in: list of invalid values for this column.
                                               it could also be a callable without any inputs.
                                               if a non-callable is provided and the column is
                                               not a primary key and also column name is provided,
                                               it will result in check constraint generation on
                                               database. otherwise it will be ignored and could
                                               be used in validators.
                                               defaults to None if not provided.

        :note check_in, check_not_in: only one of these options could be provided.
                                      otherwise it raises an error.

        :keyword bool validated: specifies that an automatic validator for this column
                                 must be registered, that is usable through validator
                                 services. defaults to False if not provided.
        """

        self.allow_read = kwargs.pop('allow_read', True)
        self.allow_write = kwargs.pop('allow_write', True)
        self.min_value = kwargs.pop('min_value', None)
        self.max_value = kwargs.pop('max_value', None)
        self.check_in = kwargs.pop('check_in', None)
        self.check_not_in = kwargs.pop('check_not_in', None)
        self.validated = kwargs.pop('validated', False)

        super().__init__(*args, **kwargs)

    def _get_schema_items(self):
        """
        gets required schema items for this column.

        it will generate required check constraints, sequences and ...

        :raises CheckConstraintConflictError: check constraint conflict error.
        :raises InvalidCheckConstraintError: invalid check constraint error.

        :rtype: list
        """

        custom_items = self._get_custom_schema_items()
        if self.name in (None, '') or self.type is None:
            return custom_items

        if self.primary_key is not True:
            minimum = None
            maximum = None
            if self.min_value is not None and not callable(self.min_value):
                minimum = self.min_value

            if self.max_value is not None and not callable(self.max_value):
                maximum = self.max_value

            if minimum is not None or maximum is not None:
                range_constraint = range_check_constraint(self.name,
                                                          min_value=minimum,
                                                          max_value=maximum)
                custom_items.append(range_constraint)

            if self.check_in is not None and self.check_not_in is not None:
                raise CheckConstraintConflictError('Both "check_in" and "check_not_in" could '
                                                   'not be provided at the same time.')

            is_check_in_callable = callable(self.check_in)
            is_check_not_in_callable = callable(self.check_not_in)
            if self.check_in is not None and not is_check_in_callable \
                    and not (isinstance(self.check_in, LIST_TYPES)
                             and len(self.check_in) > 0):
                raise InvalidCheckConstraintError('Provided value for "check_in" '
                                                  'must be an iterable with at least 1 item.')

            if self.check_not_in is not None and not is_check_not_in_callable \
                    and not (isinstance(self.check_not_in, LIST_TYPES)
                             and len(self.check_not_in) > 0):
                raise InvalidCheckConstraintError('Provided value for "check_not_in" '
                                                  'must be an iterable with at least 1 item.')

            valid_list = None
            invalid_list = None
            if self.check_in is not None and not is_check_in_callable:
                valid_list = self.check_in
            elif self.check_not_in is not None and not is_check_not_in_callable:
                invalid_list = self.check_not_in

            if valid_list is not None:
                constraint = check_constraint(self.name, valid_list)
                custom_items.append(constraint)
            elif invalid_list is not None:
                constraint = check_constraint(self.name, invalid_list, use_in=False)
                custom_items.append(constraint)

        return custom_items

    def _get_custom_schema_items(self):
        """
        gets custom schema items for this column.

        it will generate required check constraints, sequences and ...
        this method is intended to be overridden in subclasses.

        :rtype: list
        """

        return []

    def _init_items(self, *args):
        """
        initializes the list of required items for this column.
        """

        custom_items = self._get_schema_items()
        args = list(args)
        args.extend(custom_items)
        super()._init_items(*args)

    def _real_name(self):
        """
        gets the column's real name.

        :rtype: str
        """

        for column in self.base_columns:
            return column.name

    def _extract_name_and_type(self, args, kwargs):
        """
        extracts name and type parameters from given inputs and removes them from inputs.

        it returns None as each parameter that is not given.
        it is implemented to be used by custom column subclasses.

        :param list args: column positional arguments.
        :param dict kwargs: column keyword arguments.

        :raises ArgumentError: argument error.

        :returns: tuple[str name, CoreColumn type]
        :rtype: tuple[str, CoreColumn]
        """

        name = kwargs.pop('name', None)
        type_ = kwargs.pop('type_', None)
        if len(args) > 0:
            if isinstance(args[0], util.string_types):
                if name is not None:
                    raise ArgumentError('May not pass name positionally and as a keyword.')
                name = args.pop(0)

        if len(args) > 0:
            column_type = args[0]
            if hasattr(column_type, "_sqla_type"):
                if type_ is not None:
                    raise ArgumentError('May not pass type_ positionally and as a keyword.')
                type_ = args.pop(0)

        return name, type_

    def get_python_type(self, type_=None):
        """
        gets the python equivalent type of this column's type or given type.

        the returned value is a tuple of two items. first item is the collection
        type if the type is a collection, otherwise its None. the second item is
        the type of values of this column.

        it may return (None, None) if type could not be determined.
        this method is only used on server startup for registering auto validators.

        :param TypeEngine type_: sqlalchemy type instance to get its equivalent python
                                 type. if not provided, the column's type will be used.

        :returns: tuple[type collection_type, type python_type]
        :rtype: tuple[type, type]
        """

        if type_ is None:
            type_ = self.type

        python_type = None
        collection_type = None
        try:
            python_type = type_.python_type
        except (NotImplementedError, AttributeError):
            if isinstance(type_, (mssql_types.BIT, sybase_types.BIT)):
                return None, bool
            if isinstance(type_, (pg_types.UUID,
                                  mssql_types.UNIQUEIDENTIFIER,
                                  sybase_types.UNIQUEIDENTIFIER)):
                return None, UUID

        if python_type is list:
            collection_type = python_type
            python_type = None
            if isinstance(type_, ARRAY) and type_.dimensions == 1:
                __, python_type = self.get_python_type(type_.item_type)

        return collection_type, python_type

    @cached_property
    def fullname(self):
        """
        gets the column's fullname.

        fullname is made up of `table_fullname.column_name`.
        if the column has no table, it only returns the `column_name`.

        :rtype: str
        """

        if self.table is None:
            return self._real_name()

        for single_column in self.table.columns:
            if single_column.name == self.name:
                for base_column in single_column.base_columns:
                    return '{table}.{column}'.format(table=base_column.table.fullname,
                                                     column=self._real_name())

    @property
    def is_foreign_key(self):
        """
        gets a value indicating that this column is a foreign key.

        :rtype: bool
        """

        return len(self.foreign_keys) > 0



# -*- coding: utf-8 -*-
"""
permission exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class PermissionManagerException(CoreException):
    """
    permission manager exception.
    """
    pass


class PermissionManagerBusinessException(CoreBusinessException,
                                         PermissionManagerException):
    """
    permission manager business exception.
    """
    pass


class InvalidPermissionTypeError(PermissionManagerException):
    """
    invalid permission type error.
    """
    pass


class DuplicatedPermissionError(PermissionManagerException):
    """
    duplicated permission error.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.cli import CLIPackage
from pyrin.cli.manager import CLIManager


@component(CLIPackage.COMPONENT_NAME)
class CLIComponent(Component, CLIManager):
    """
    cli component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
string normalizer component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.utilities.string.normalizer import StringNormalizerPackage
from pyrin.utilities.string.normalizer.manager import StringNormalizerManager


@component(StringNormalizerPackage.COMPONENT_NAME)
class StringNormalizerComponent(Component, StringNormalizerManager):
    """
    string normalizer component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
core enumerations module.
"""

import inspect

from copy import deepcopy
from collections import OrderedDict


class EnumMember:
    """
    enum member class.

    if you need localizable values for your enumerations, you could define
    enum values of this type instead of primitive types.

    for example:

    class Animal(CoreEnum):
        RABBIT = EnumMember(1, 'Rabbit')
        LION = EnumMember(2, 'Lion')
        UNKNOWN = 3

    animal_1 = Animal.RABBIT --> this returns the value of rabbit which is 1.
    animal_1_name = Animal(Animal.RABBIT) --> this returns the name of rabbit
    which is `Rabbit`.
    animal_2 = Animal.UNKNOWN --> this returns the value of unknown which is 3.
    animal_2_name = Animal(3) --> this returns the name of unknown which is `UNKNOWN`.
    animal_5 = Animal(5, silent=True) --> this returns None.
    animal_6 = Animal(6) --> this raises an attribute error.

    as you could see, you could also define enum values with primitive types if needed.
    but they won't have localizable names.
    """

    def __init__(self, value, name):
        """
        initializes an instance of EnumMember.

        :param value: enum value.
        :param str name: enum name.
        """

        self._value = value
        self._name = name

    def __str__(self):
        """
        gets the string representation of this enum member.

        :rtype: str
        """

        return self.name

    def __repr__(self):
        """
        gets the string representation of this enum member.

        :rtype: str
        """

        return '[{value}:{name}]'.format(value=self.value, name=self.name)

    def __eq__(self, other):
        """
        gets a value indicating that this enum value is equal to provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value == other.value

        return self.value == other

    def __ne__(self, other):
        """
        gets a value indicating that this enum value is not equal to provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value != other.value

        return self.value != other

    def __ge__(self, other):
        """
        implements `>=` operator to compare current enum member with provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value >= other.value

        return self.value >= other

    def __le__(self, other):
        """
        implements `<=` operator to compare current enum member with provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value <= other.value

        return self.value <= other

    def __gt__(self, other):
        """
        implements `>` operator to compare current enum member with provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value > other.value

        return self.value > other

    def __lt__(self, other):
        """
        implements `<` operator to compare current enum member with provided value.

        :param EnumMember | object other: other object to be compared
                                          to this enum member.

        :rtype: bool
        """

        if isinstance(other, EnumMember):
            return self.value < other.value

        return self.value < other

    def __hash__(self):
        """
        gets the hash of this enum member.

        :rtype: int
        """

        return hash(self.value)

    @property
    def value(self):
        """
        gets this enum member's value.

        :returns: object
        """

        return self._value

    @property
    def name(self):
        """
        gets this enum member's name.

        :rtype: str
        """

        return self._name


class CoreEnumMeta(type):
    """
    core enum metaclass.
    """

    # each enum class will gain it's own attributes to save names and values of
    # it's members. the below attributes will always remain None and never get populated.
    __enum_names = None
    __enum_values = None
    __enum_dict = None

    @staticmethod
    def __new__(mcs, *args, **kwargs):
        """
        creates and returns a new class.

        note that this method will only get called once per each enum subclass.

        :rtype: type[CoreEnumMeta]
        """

        mcs._check_duplicates(args[0], args[2])
        return super().__new__(mcs, *args, **kwargs)

    def __getattribute__(cls, name):
        """
        gets an attribute of this class with given name.

        it returns the enum value, even if it is of `EnumMember` type.

        :param str name: attribute name to be get.

        :returns: object
        """

        member = type.__getattribute__(cls, name)
        # we should not call 'get_pure_value' here because it results in recursion error.
        if isinstance(member, EnumMember):
            member = member.value

        return member

    def __contains__(cls, member):
        """
        gets a value indicating that given input exists in the enumeration values.

        this method is overridden to be able to check
        for existence with `in` keyword. for example:
        has_value = 'value' in SomeEnum

        :param EnumMember | object member: value to be checked for existence.

        :rtype: bool
        """

        member = cls.get_pure_value(member)
        return member in cls.values()

    def __to_dict(cls):
        """
        gets a dict containing all values of this enum and their names.

        :returns: dict(object value, str name)
        :rtype: dict
        """

        if cls.__enum_dict is None:
            cls._populate_members()

        return cls.__enum_dict

    @staticmethod
    def _check_duplicates(enum, attrs):
        """
        checks that is there any enum members with the same value.

        if it does, it raises an error.

        :param str enum: enum name.
        :param dict attrs: all attributes of this class.

        :raise ValueError: value error.
        """

        values = []
        for name, value in attrs.items():
            if CoreEnumMeta._is_enumeration(name, value):
                pure_value = CoreEnumMeta.get_pure_value(value)
                if pure_value in values:
                    raise ValueError('Enumeration [{enum}] has multiple '
                                     'members with value [{value}].'
                                     .format(enum=enum, value=pure_value))

                values.append(pure_value)
                continue

    @staticmethod
    def _is_enumeration(name, value):
        """
        gets a value indicating that given name belongs to an enum member.

        enum members must not start or end with underscores `_`.
        callable values are also ignored.
        all `EnumMember` instances will be accepted without considering their naming style.

        :param str name: name to be checked.
        :param EnumMember | object value: value to be checked.

        :rtype: bool
        """

        if isinstance(value, EnumMember):
            return True

        if name is None or value is None or callable(value) or \
                inspect.isfunction(value) or inspect.ismethod(value) or \
                inspect.ismethoddescriptor(value) or inspect.isclass(value):
            return False

        return not name.startswith('_') and not name.endswith('_')

    @staticmethod
    def get_pure_value(member):
        """
        gets the pure value of given value.

        if the value is an instance of `EnumMember` it returns the `value` of it.
        otherwise it returns the same input.

        :param EnumMember | object member: value to get its pure value.

        :returns: object
        """

        if isinstance(member, EnumMember):
            return member.value

        return member

    def _populate_members(cls):
        """
        populates this enum's names and values and dictionary.
        """

        names = []
        values = []
        dictionary = OrderedDict()
        for name in cls.__dict__:
            item = cls.__dict__.get(name)
            if cls._is_enumeration(name, item):
                if isinstance(item, EnumMember):
                    name = item.name
                    item = item.value
                names.append(name)
                values.append(item)
                dictionary[item] = name

        cls.__enum_names = tuple(names)
        cls.__enum_values = tuple(values)
        cls.__enum_dict = dictionary

    def values(cls):
        """
        gets a tuple containing all values in the enumeration.

        :rtype: tuple
        """

        if cls.__enum_values is None:
            cls._populate_members()

        return cls.__enum_values

    def names(cls):
        """
        gets a tuple containing all names in the enumeration.

        :rtype: tuple
        """

        if cls.__enum_names is None:
            cls._populate_members()

        return cls.__enum_names

    def to_dict(cls):
        """
        gets a dict containing all values of this enum and their names.

        :returns: dict(object value, str name)
        :rtype: dict
        """

        return deepcopy(cls.__to_dict())

    def str(cls, value):
        """
        gets the name of given value in enumeration.

        note that if the value is not an `EnumMember` instance, it does not have
        a name and the name of it's attribute will be returned instead.
        if the value does not exist in enumeration, it raises an error.

        :param EnumMember | object value: value to get its name.

        :raises AttributeError: attribute error.

        :rtype: str
        """

        value = cls.get_pure_value(value)
        name = cls.__to_dict().get(value)
        if name is None:
            raise AttributeError('Enumeration [{enum}] does not have '
                                 'a member with value [{value}].'
                                 .format(enum=cls.__name__, value=value))

        return name


class CoreEnum(metaclass=CoreEnumMeta):
    """
    core enum class.

    all application enumerations must inherit from this class.
    """

    @staticmethod
    def __new__(cls, value, silent=False):
        """
        gets the name of given value in enumeration.

        note that if the value is not an `EnumMember` instance, it does not have
        a name and the name of it's attribute will be returned instead.
        if the value does not exist in enumeration, it raises an error.

        example usage:

        class CarEnum(CoreEnum):
            MERCEDES = EnumMember(1, 'Mercedes')
            BMW = 2

        name1 = CarEnum(1) --> `Mercedes`
        name2 = CarEnum(CarEnum.MERCEDES) --> `Mercedes`
        name3 = CarEnum(2000, silent=True) --> None
        name4 = CarEnum(2000) --> raises AttributeError
        name5 = CarEnum(2) --> `BMW`

        :param EnumMember | object value: value to get its name.

        :param bool silent: specifies that if value does not exist in
                            enumeration, does not raise an error and
                            return None instead. defaults to False.

        :raises AttributeError: attribute error.

        :rtype: str
        """

        if silent is True:
            return cls.try_str(value)

        return cls.str(value)

    @classmethod
    def contains(cls, value):
        """
        gets a value indicating that given input exists in the enumeration values.

        :param EnumMember | object value: value to be checked for existence.

        :rtype: bool
        """

        value = cls.get_pure_value(value)
        return value in cls.values()

    @classmethod
    def try_str(cls, value):
        """
        gets the name of given value in enumeration.

        note that if the value is not an `EnumMember` instance, it does not have
        a name and the name of it's attribute will be returned instead.
        if the value does not exist in enumeration, it returns None.

        :param EnumMember | object value: value to get its name.

        :rtype: str
        """

        try:
            return cls.str(value)
        except AttributeError:
            return None


class HTTPMethodEnum(CoreEnum):
    """
    http method enum.
    """

    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    HEAD = 'HEAD'
    TRACE = 'TRACE'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    COPY = 'COPY'
    LINK = 'LINK'
    UNLINK = 'UNLINK'
    PURGE = 'PURGE'
    VIEW = 'VIEW'


class InformationResponseCodeEnum(CoreEnum):
    """
    http information response code enum.
    """

    # this interim response indicates that everything so far is OK and that
    # the client should continue with the request or ignore it if it is already finished.
    CONTINUE = 100

    # this code is sent in response to an upgrade request header by the client,
    # and indicates the protocol the server is switching to.
    SWITCHING_PROTOCOL = 101

    # this code indicates that the server has received and
    # is processing the request, but no response is available yet.
    PROCESSING = 102

    # this status code is primarily intended to be used with the Link header to allow
    # the user agent to start preloading resources while the server is still preparing a response.
    EARLY_HINTS = 103


class SuccessfulResponseCodeEnum(CoreEnum):
    """
    http successful response code enum.
    """

    # the request has succeeded. the meaning of a success varies depending on the http method:
    # GET: the resource has been fetched and is transmitted in the message body.
    # HEAD: the entity headers are in the message body.
    # PUT/POST: the resource describing the result of the action is transmitted in the message
    # body.
    # TRACE: the message body contains the request message as received by the server.
    OK = 200

    # the request has succeeded and a new resource has been created as a result of it.
    # this is typically the response sent after a POST request, or after some PUT requests.
    CREATED = 201

    # the request has been received but not yet acted upon. it is non-committal,
    # meaning that there is no way in http to later send an asynchronous response
    # indicating the outcome of processing the request. it is intended for cases where
    # another process or server handles the request, or for batch processing.
    ACCEPTED = 202

    # this response code means returned meta-information set is not exact set as
    # available from the origin server, but collected from a local or a third party copy.
    # except this condition, 200 OK response should be preferred instead of this response.
    NON_AUTHORITATIVE_INFORMATION = 203

    # there is no content to send for this request, but the headers may be useful.
    # the user-agent may update its cached headers for this resource with the new ones.
    NO_CONTENT = 204

    # this response code is sent after accomplishing request
    # to tell user agent reset document view which sent this request.
    RESET_CONTENT = 205

    # this response code is used because of range header sent
    # by the client to separate download into multiple streams.
    PARTIAL_CONTENT = 206

    # a multi-status response conveys information about multiple resources
    # in situations where multiple status codes might be appropriate.
    MULTI_STATUS = 207

    # used inside a DAV: propstat response element to avoid enumerating
    # the internal members of multiple bindings to the same collection repeatedly.
    ALREADY_REPORTED = 208

    # the server has fulfilled a GET request for the resource,
    # and the response is a representation of the result of
    # one or more instance-manipulations applied to the current instance.
    IM_USED = 226


class RedirectionResponseCodeEnum(CoreEnum):
    """
    http redirection response code enum.
    """

    # the request has more than one possible response. the user-agent or user should
    # choose one of them. there is no standardized way of choosing one of the responses.
    MULTIPLE_CHOICE = 300

    # this response code means that the URI of the requested resource
    # has been changed permanently. Probably, the new URI would be given in the response.
    MOVED_PERMANENTLY = 301

    # this response code means that the URI of requested resource has been changed temporarily.
    # new changes in the URI might be made in the future.
    # therefore, this same URI should be used by the client in future requests.
    FOUND = 302

    # the server sent this response to direct the client to get the
    # requested resource at another URI with a GET request.
    SEE_OTHER = 303

    # this is used for caching purposes. it tells the client that the response has not
    # been modified, so the client can continue to use the same cached version of the response.
    NOT_MODIFIED = 304

    # was defined in a previous version of the http specification to indicate that a
    # requested response must be accessed by a proxy. it has been deprecated due to
    # security concerns regarding in-band configuration of a proxy.
    USE_PROXY = 305

    # this response code is no longer used, it is just reserved currently.
    # it was used in a previous version of the http 1.1 specification.
    UNUSED = 306

    # the server sends this response to direct the client to get the requested resource
    # at another URI with same method that was used in the prior request.
    # this has the same semantics as the 302 Found http response code, with the exception
    # that the user agent must not change the http method used: if a POST was used in
    # the first request, a POST must be used in the second request.
    TEMPORARY_REDIRECT = 307

    # this means that the resource is now permanently located at another URI,
    # specified by the Location: http response header. this has the same semantics
    # as the 301 moved permanently http response code, with the exception that
    # the user agent must not change the http method used: if a POST was used in
    # the first request, a POST must be used in the second request.
    PERMANENT_REDIRECT = 308


class ClientErrorResponseCodeEnum(CoreEnum):
    """
    http client error response code enum.
    """

    # this response means that server could not understand the request due to invalid syntax.
    BAD_REQUEST = 400

    # although the http standard specifies "unauthorized", semantically
    # this response means "unauthenticated". that is, the client must authenticate
    # itself to get the requested response.
    UNAUTHORIZED = 401

    # this response code is reserved for future use. initial aim for creating
    # this code was using it for digital payment systems however this is not used currently.
    PAYMENT_REQUIRED = 402

    # The client does not have access rights to the content, i.e. they are unauthorized,
    # so server is rejecting to give proper response. unlike 401, the client's identity
    # is known to the server.
    FORBIDDEN = 403

    # the server can not find requested resource. in the browser, this means the url
    # is not recognized. in an API, this can also mean that the endpoint is valid but the
    # resource itself does not exist. servers may also send this response instead of 403
    # to hide the existence of a resource from an unauthorized client.
    # this response code is probably the most famous one due to its frequent occurrence on the web.
    NOT_FOUND = 404

    # the request method is known by the server but has been disabled and cannot be used.
    # for example, an API may forbid DELETE-ing a resource. The two mandatory methods,
    # GET and HEAD, must never be disabled and should not return this error code.
    METHOD_NOT_ALLOWED = 405

    # this response is sent when the web server, after performing server-driven content
    # negotiation, doesn't find any content following the criteria given by the user agent.
    NOT_ACCEPTABLE = 406

    # this is similar to 401 but authentication is needed to be done by a proxy.
    PROXY_AUTHENTICATION_REQUIRED = 407

    # this response is sent on an idle connection by some servers, even without any previous
    # request by the client. it means that the server would like to shut down this unused
    # connection. this response is used much more since some browsers, like Chrome,
    # Firefox 27+, or IE9, use http pre-connection mechanisms to speed up surfing.
    # also note that some servers merely shut down the connection without sending this message.
    REQUEST_TIMEOUT = 408

    # this response is sent when a request conflicts with the current state of the server.
    CONFLICT = 409

    # this response would be sent when the requested content has been permanently deleted
    # from server, with no forwarding address. clients are expected to remove their caches
    # and links to the resource. The http specification intends this status code to be used
    # for "limited-time, promotional services". APIs should not feel compelled to indicate
    # resources that have been deleted with this status code.
    GONE = 410

    # server rejected the request because the Content-Length header field
    # is not defined and the server requires it.
    LENGTH_REQUIRED = 411

    # the client has indicated preconditions in its headers which the server does not meet.
    PRECONDITION_FAILED = 412

    # request entity is larger than limits defined by server; the server might
    # close the connection or return an Retry-After header field.
    PAYLOAD_TOO_LARGE = 413

    # the URI requested by the client is longer than the server is willing to interpret.
    URI_TOO_LONG = 414

    # the media format of the requested data is not supported by the server,
    # so the server is rejecting the request.
    UNSUPPORTED_MEDIA_TYPE = 415

    # the range specified by the Range header field in the request can't be fulfilled;
    # it's possible that the range is outside the size of the target URI's data.
    REQUESTED_RANGE_NOT_SATISFIABLE = 416

    # this response code means the expectation indicated by the
    # expect request header field can't be met by the server.
    EXPECTATION_FAILED = 417

    # the server refuses the attempt to brew coffee with a teapot.
    IM_A_TEAPOT = 418

    # the request was directed at a server that is not able to produce a response.
    # this can be sent by a server that is not configured to produce responses for
    # the combination of scheme and authority that are included in the request URI.
    MISDIRECTED_REQUEST = 421

    # the request was well-formed but was unable to be followed due to semantic errors.
    UNPROCESSABLE_ENTITY = 422

    # the resource that is being accessed is locked.
    LOCKED = 423

    # the request failed due to failure of a previous request.
    FAILED_DEPENDENCY = 424

    # indicates that the server is unwilling to risk processing a request that might be replayed.
    TOO_EARLY = 425

    # the server refuses to perform the request using the current protocol but might be
    # willing to do so after the client upgrades to a different protocol.
    # the server sends an Upgrade header in a 426 response to indicate the required protocol(s).
    UPGRADE_REQUIRED = 426

    # the origin server requires the request to be conditional. intended to prevent
    # the 'lost update' problem, where a client GETs a resource's state, modifies it,
    # and PUTs it back to the server, when meanwhile a third party has modified the state
    # on the server, leading to a conflict.
    PRECONDITION_REQUIRED = 428

    # the user has sent too many requests in a given amount of time ("rate limiting").
    TOO_MANY_REQUESTS = 429

    # the server is unwilling to process the request because its header fields are too large.
    # the request may be resubmitted after reducing the size of the request header fields.
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431

    # the user requests an illegal resource, such as a web page censored by a government.
    UNAVAILABLE_FOR_LEGAL_REASONS = 451


class ServerErrorResponseCodeEnum(CoreEnum):
    """
    http server error response code enum.
    """

    # the server has encountered a situation it doesn't know how to handle.
    INTERNAL_SERVER_ERROR = 500

    # the request method is not supported by the server and cannot be handled.
    # the only methods that servers are required to support (and therefore that must not
    # return this code) are GET and HEAD.
    NOT_IMPLEMENTED = 501

    # this error response means that the server, while working as a gateway to get
    # a response needed to handle the request, got an invalid response.
    BAD_GATEWAY = 502

    # the server is not ready to handle the request. common causes are a server that
    # is down for maintenance or that is overloaded. Note that together with this response,
    # a user-friendly page explaining the problem should be sent. This responses should be
    # used for temporary conditions and the Retry-After: http header should, if possible,
    # contain the estimated time before the recovery of the service. the webmaster
    # must also take care about the caching-related headers that are sent along with
    # this response, as these temporary condition responses should usually not be cached.
    SERVICE_UNAVAILABLE = 503

    # this error response is given when the server is acting
    # as a gateway and cannot get a response in time.
    GATEWAY_TIMEOUT = 504

    # the http version used in the request is not supported by the server.
    HTTP_VERSION_NOT_SUPPORTED = 505

    # the server has an internal configuration error: transparent content negotiation
    # for the request results in a circular reference.
    VARIANT_ALSO_NEGOTIATES = 506

    # the server has an internal configuration error: the chosen variant resource
    # is configured to engage in transparent content negotiation itself,
    # and is therefore not a proper end point in the negotiation process.
    INSUFFICIENT_STORAGE = 507

    # the server detected an infinite loop while processing the request.
    LOOP_DETECTED = 508

    # further extensions to the request are required for the server to fulfill it.
    NOT_EXTENDED = 510

    # the 511 status code indicates that the client needs to authenticate to gain network access.
    NETWORK_AUTHENTICATION_REQUIRED = 511



# -*- coding: utf-8 -*-
"""
validator auto services module.
"""

from pyrin.application.services import get_component
from pyrin.validator.auto import ValidatorAutoPackage


def register_auto_validator(domain, field, **options):
    """
    register required auto validator for given field.

    :param BaseEntity domain: entity type that this field is related to.
    :param InstrumentedAttribute field: field instance.
    """

    return get_component(ValidatorAutoPackage.COMPONENT_NAME).register_auto_validator(
        domain, field, **options)


def register_auto_validators():
    """
    registers all auto validators.

    it returns the count of registered auto validators.

    :rtype: int
    """

    return get_component(ValidatorAutoPackage.COMPONENT_NAME).register_auto_validators()



# -*- coding: utf-8 -*-
"""
orm session base module.
"""

from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import TextClause

import pyrin.database.services as database_services
import pyrin.database.orm.sql.extractor.services as extractor_services

from pyrin.core.globals import _
from pyrin.database.exceptions import InvalidDatabaseBindError
from pyrin.database.orm.session.exceptions import TransientSQLExpressionRequiredError


class CoreSession(Session):
    """
    core session class.

    all application sessions must be an instance this.
    """

    # these keywords are forbidden when 'execute' method is called
    # with 'transient=True' and a raw sql is given.
    NON_TRANSIENT_KEYWORDS = ['commit', 'flush', 'rollback', 'alter ', 'create ', 'drop ']

    def execute(self, clause, params=None, mapper=None, bind=None, **kw):
        """
        executes a sql expression construct or string statement.

        it will be executed within the current transaction.
        the current transaction will be resolved using the given
        entity class or object or bind name.

        this method is a wrapper for `sqlalchemy.orm.session.execute()` method
        which wraps it to bind to correct engine on different situations.
        for full documentation see the `sqlalchemy.orm.session.execute()` docs.

        :param Executable | str clause: an executable statement `Executable`
                                        expression such as `expression.select`
                                        or string sql statement to be executed.

        :param dict | list[dict] params: optional dictionary, or list of
                                         dictionaries, containing bound
                                         parameter values. if a single dictionary,
                                         single-row execution occurs. if a list of
                                         dictionaries, an `executemany' will be invoked.
                                         the keys in each dictionary must correspond
                                         to parameter names present in the statement.

        :param BaseEntity mapper: entity class or object which the given clause
                                  must be executed on its related engine.
                                  this argument takes precedence over
                                  `clause` when locating a bind.

        :param Engine bind: optional engine to be used as the bind. if
                            this engine is already involved in an ongoing transaction,
                            that connection will be used. this argument takes
                            precedence over `mapper` and `clause` and `bind_name` when
                            locating a bind.

        :keyword kw: additional keyword arguments are sent to `session.get_bind()`
                     to allow extensibility of `bind` schemes.

        :keyword str bind_name: database bind name from `database.binds.ini`
                                file on which given clause must be executed.
                                this keyword argument precedence over `mapper` and
                                `clause` when locating a bind.

        :keyword bool transient: specifies that raw sql expressions must not affect
                                 the database. if set to True and a raw sql is
                                 given, it could not contain `NON_TRANSIENT_KEYWORDS`.
                                 defaults to False if not provided.

        :raises InvalidDatabaseBindError: invalid database bind error.
        :raises TransientSQLExpressionRequiredError: transient sql expression required error.

        :returns: results of the statement execution.
        :rtype: ResultProxy
        """

        bind_name = kw.pop('bind_name', None)
        if bind is None and bind_name is not None:
            bind = database_services.get_bounded_engines().get(bind_name, None)
            if bind is None:
                raise InvalidDatabaseBindError('Database bind name [{bind_name}] '
                                               'is not available in database.binds '
                                               'config store.'
                                               .format(bind_name=bind_name))

        transient = kw.pop('transient', False)
        if transient is True and isinstance(clause, (str, TextClause)):
            raw_sql = None
            if isinstance(clause, TextClause):
                raw_sql = clause.text.lower()
            else:
                raw_sql = clause.lower()

            for item in self.NON_TRANSIENT_KEYWORDS:
                if item in raw_sql:
                    raise TransientSQLExpressionRequiredError(_('Transient sql expressions '
                                                                'should not contain [{keyword}] '
                                                                'keyword.')
                                                              .format(keyword=item.strip()))

        return super().execute(clause, params, mapper, bind, **kw)

    def get_bind(self, mapper=None, clause=None, **kw):
        """
        returns a `bind` to which this `CoreSession` is bounded.

        this method extends the `get_bind() method of `Session` by
        trying to find bounded engine to tables of a raw string query.

        the `bind` is usually an instance of `Engine`,
        except in the case where the `CoreSession` has been
        explicitly bound directly to a `Connection`.

        for a multiply-bound or unbound `CoreSession`, the
        `mapper` or `clause` arguments are used to determine the
        appropriate bind to return.

        note that the 'mapper' argument is usually present
        when `CoreSession.get_bind` is called via an orm
        operation such as a `Session.query`, each
        individual INSERT/UPDATE/DELETE operation within a
        `Session.flush`, call, etc.

        The order of resolution is:

        1. if mapper given and session.binds is present,
           locate a bind based first on the mapper in use, then
           on the mapped class in use, then on any base classes that are
           present in the `__mro__` of the mapped class, from more specific
           superclasses to more general.
        2. if clause given and session.binds is present,
           locate a bind based on `Table` objects
           found in the given clause present in session.binds.
        3. if session.bind is present, return that.
        4. if clause given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the clause.
        5. if mapper given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the `Table` or other
           selectable to which the mapper is mapped.
        6. no bind can be found, `sqlalchemy.exc.UnboundExecutionError`
           is raised.

        :param BaseEntity mapper: a BaseEntity instance or class. the bind can be derived from
                                  a `Mapper` first by consulting the `binds` map associated
                                  with this `Session`, and secondly by consulting the
                                  `Metadata` associated with the `Table` to which the
                                  `Mapper` is mapped for a bind.

        :param ClauseElement | str clause: a raw string query or a `ClauseElement`
                                           (i.e. `~.sql.expression.select`,
                                           `~.sql.expression.text`, etc.). if the
                                           `Mapper` argument is not present or could not
                                           produce a bind, the given expression construct
                                           will be searched for a bound element, typically
                                           a `Table` associated with `Metadata`.

        :rtype: Engine
        """

        if mapper is None and isinstance(clause, (str, TextClause)):
            tables = extractor_services.find_table_names(clause)
            if len(tables) > 0:
                return database_services.get_table_engine(tables[0])

        return super().get_bind(mapper, clause=clause)



# -*- coding: utf-8 -*-
"""
database audit hooks module.
"""

import pyrin.database.audit.services as database_audit_services

from pyrin.audit.decorators import audit_hook
from pyrin.audit.hooks import AuditHookBase


@audit_hook()
class AuditHook(AuditHookBase):
    """
    audit hook class.
    """

    audit_name = 'database'

    def inspect(self, **options):
        """
        this method will be called to inspect the status of a package or resource.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        return database_audit_services.inspect(**options)



# -*- coding: utf-8 -*-
"""
database audit component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.database.audit import DatabaseAuditPackage
from pyrin.database.audit.manager import DatabaseAuditManager


@component(DatabaseAuditPackage.COMPONENT_NAME)
class DatabaseAuditComponent(Component, DatabaseAuditManager):
    """
    database audit component class.
    """
    pass
